#!/usr/bin/python3
# Copyright (C) 2019, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.


import asyncio
import argparse
import code
import collections.abc
import concurrent.futures
import shlex
import re
import time
import pkg_resources
import os.path
import readline
import rlcompleter
import sys
import traceback
import carthage
import carthage.utils
from carthage import base_injector, ConfigLayout
from carthage.dependency_injection import *

__all__ = []

subcommands_re = re.compile(r'^\s*!\s*(\S.*)\s*$')


class EventMonitor:

    def __init__(self, injector):
        injector.add_event_listener(InjectionKey(Injector), "dependency_progress", self.progress_cb)
        injector.add_event_listener(InjectionKey(Injector), "dependency_final", self.final_cb)
        self.injector = injector

    def close(self):
        self.injector.remove_event_listener(InjectionKey(Injector), "dependency_progress", self.progress_cb)
        self.remove_event_listener(InjectionKey(Injector), "dependency_final", self.final_cb)

    def progress_cb(self, target, **kwargs):
        print(f'Progress: {target}: {target.provider.provider}')

    def final_cb(self, target, **kwargs):
        print(f'Resolved: {target}: {target.get_value(False)}')


def attach_event_monitor_to_console(injector):
    return EventMonitor(injector)


class CarthageConsole(code.InteractiveConsole):

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--no-console', action='store_false',
                            dest='console', default=True,
                            help="Do not run the console")
        parser.add_argument('--console', action='store_true',
                            dest='console', default=True,
                            help="Run the console")
        parser.add_argument('--rcfile',
                            metavar="file",
                            default="~/.carthagerc",
                            help="Python code to run in Carthage console")

    def process_arguments(self, args):
        if args.rcfile:
            try:
                with open(os.path.expanduser(args.rcfile), "rt") as f:
                    exec(f.read(), self.locals)
            except FileNotFoundError:
                pass

    @staticmethod
    def noop():
        # dummy function used to start loop when called from another thread
        pass

    @staticmethod
    def default_locals():
        from carthage.dependency_injection.introspection import instantiation_roots
        return {
            'injector': base_injector,
            'instantiation_roots': instantiation_roots,
            'ainjector': base_injector(AsyncInjector),
            'loop': asyncio.get_event_loop(),
            'attach_monitor': attach_event_monitor_to_console,
            'config': base_injector(ConfigLayout)
        }

    def exec_resource(self, pkg, resource):
        res_str = pkg_resources.resource_string(pkg, resource)
        exec(compile(res_str, resource, mode="exec"), self.locals)

    def __init__(self, locals=None, extra_locals=None,
                 history_file="~/.carthage_history"):
        if locals is None:
            locals = self.default_locals()
        if extra_locals is not None:
            locals = {**locals, **extra_locals}
        super().__init__(locals=locals)
        self.locals = locals
        self.orig_completer = None
        self.orig_displayhook = None
        self.completed_keys = []
        self.history = {}
        self.history_num = 0
        self.loop = asyncio.get_event_loop()
        if 'h' in self.locals and self.locals['h'] is not self.history:
            raise NotImplementedError(
                "When replacing `h', it is unclear whether you want to link to the async history object or disable it.")
        self.locals['h'] = self.history
        self.subcommands = None
        self.subcommands_parser = None
        if history_file:
            history_file = os.path.expanduser(history_file)
            try:
                readline.read_history_file(history_file)
            except FileNotFoundError:
                pass
        self.history_file = history_file

    def interact(self, *args, **kwargs):
        kwargs  = dict(kwargs)
        if self.subcommands and 'banner' not in kwargs:
            kwargs['banner'] = "Carthage console.  Type Python expressions or use !help for Carthage commands."
        try:
            asyncio.get_event_loop()
        except Exception:  # Running in thread without loop
            asyncio.set_event_loop(self.loop)
        self.orig_completer = readline.get_completer()
        self.orig_displayhook = sys.displayhook

        sys.displayhook = self.displayhook
        completer = rlcompleter.Completer(self.locals)
        readline.set_completer(completer.complete)
        readline.parse_and_bind('tab: complete')

        try:
            super().interact(*args, **kwargs)
        finally:
            readline.set_completer(self.orig_completer)
            sys.displayhook = self.orig_displayhook
            if self.history_file:
                readline.write_history_file(self.history_file)

    def raw_input(self, *args, **kwargs):
        for k in self.completed_keys:
            print(f'[{k}]: {self.history[k]}')
        self.completed_keys.clear()
        return super().raw_input(*args, **kwargs)

    def displayhook(self, obj):

        def future_callback(f):
            del self.history[num]
            try:
                self.history[num] = f.result()
                self.completed_keys.append(num)
            except Exception as e:
                print(f'[{num}]-> exception')
                traceback.print_exc()

        if isinstance(obj, collections.abc.Coroutine):
            self.history_num += 1
            num = self.history_num
            future = asyncio.ensure_future(obj, loop=self.loop)
            future.add_done_callback(future_callback)
            print(f'[{num}]: async {obj.__name__}')
            self.history[num] = future
            self.loop.call_soon_threadsafe(CarthageConsole.noop)
        else:
            self.orig_displayhook(obj)

    async def enable_console_commands(self, ainjector):
        parser = argparse.ArgumentParser(add_help=False, exit_on_error=False)
        subparser_action = parser.add_subparsers(title='command', dest='cmd', required=True)
        subparser_action.add_parser('help', help='Print Help')
        self.subcommands = await self.setup_subcommands(ainjector, subparser_action)
        self.ainjector = ainjector
        self.subcommands_parser = parser

    async def setup_subcommands(self, ainjector, subparser_action, debug=False):
        '''
        Search through the injector for :class:`CarthageRunnerCommand` instances.  If they are available, then attach them to the parser.

        :return: Mapping of subcommand tag to subcommand instances.

        It is typical for this method to be called more than once.  For example the runner calls this both for its main parser and for a parser used within the console.
        '''
        subcommands = {}
        k = None
        v = None
        for k in ainjector.filter(
                carthage.console.CarthageRunnerCommand,
                ['name']):
            if debug:
                try:
                    v = await ainjector.get_instance_async(k)
                except InjectionFailed:
                    continue
            else:
                with injection_failed_unlogged():
                    try:
                        v = await ainjector.get_instance_async(k)
                    except InjectionFailed:
                        continue

            if await v.should_register():
                v.register(subparser_action)
            subcommands[v.name] = v
        return subcommands

    def runsource(self, source, filename):
        if self.subcommands:
            match = subcommands_re.match(source)
            if match:
                try:
                    args = self.subcommands_parser.parse_args(shlex.split(match.group(1)))
                except argparse.ArgumentError as e:
                    print(e)
                    return
                if args.cmd == 'help':
                    self.subcommands_parser.print_help()
                    return
                subcommand = self.subcommands[args.cmd]
                future = asyncio.run_coroutine_threadsafe(self.ainjector(subcommand.run, args), loop=self.loop)
                try: future.result()  #concurrent not asyncio future
                except Exception as e:
                    traceback.print_exception(e)
                return
        return super().runsource(source, filename)

    async def setup_from_plugins(self, injector=None):
        from .plugins import CarthagePlugin
        if injector is None:
            injector = base_injector
        for k, plugin in injector.filter_instantiate(CarthagePlugin, ['name']):
            if 'console_setup' in plugin.metadata:
                try:
                    exec(plugin.metadata['console_setup'], self.locals)
                    if 'async_setup' in self.locals:
                        await self.locals['async_setup']()
                except Exception:
                    print('failed to setup plugin ' + plugin.name)
                    traceback.print_exc()
                finally:
                    try:
                        del self.locals['async_setup']
                    except KeyError:
                        pass



__all__ += ['CarthageConsole']


class CarthageRunnerCommand(AsyncInjectable):

    @property
    def name(self):
        raise NotImplementedError('You must set name in a subclass')

    #: Extra arguments to be passed into .add_subparser
    subparser_kwargs = {}

    #: Does generate need to be run on the layout before this command?
    generate_required = False

    def setup_subparser(self, subparser):
        '''Generally calls add_argument a lot.'''
        raise NotImplementedError

    async def run(self, args):
        '''Called when this subcommand is selected.'''
        pass

    async def should_register(self):
        ''' Returns True if this command should be registered.  For example, start_machine might wish to confirm that the layout has Machines before registering.  This function should bd work even if for example there is no CarthageLayout.
'''
        return True

    def register(self, subparser_action):
        parser = subparser_action.add_parser(
            self.name, **self.subparser_kwargs)
        self.setup_subparser(parser)

    @classmethod
    def default_class_injection_key(cls):
        return InjectionKey(CarthageRunnerCommand, name=cls.name)


__all__ += ['CarthageRunnerCommand']


def main():

    console = CarthageConsole()
    parser = carthage.utils.carthage_main_argparser()
    CarthageConsole.add_arguments(parser)
    args = carthage.utils.carthage_main_setup(parser)
    console.process_arguments(args)
    loop = asyncio.get_event_loop()

    async def run():
        await loop.run_in_executor(None, console.interact)
    carthage.utils.carthage_main_run(run)


if __name__ == '__main__':
    main()



# Copyright (C) 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import requests
import shutil

from pyVmomi import vim, vmodl
from pathlib import Path
from carthage import *
from carthage.console import *

from .inventory import VmwareSpecifiedObject
from .folder import VmwareFolder


class DataStoreFolder(VmwareFolder, kind='datastore'):
    pass


class VmwareDataStoreCluster(VmwareSpecifiedObject, kind='datastore'):

    parent_type = DataStoreFolder

    async def do_create(self):
        self.parent.mob.CreateStoragePod(name=self.name)

    async def delete(self):
        task = self.mob.Destroy_Task()
        await carthage.vmware.utils.wait_for_task(task)


class VmwareDataStore(VmwareSpecifiedObject, kind='datastore'):

    '''
    If *local_path* is set it is either a path to a local NFS mount of the datastore, or a ``host:path`` commbination in which case the data store is accessed by scp.
'''

    parent_type = (VmwareDataStoreCluster, DataStoreFolder)

    def __init__(self, name, *, local_path=None, host=None, hosts=None, spec=None, **kwargs):

        self.spec = spec
        if (hosts is not None) and (host is not None):
            raise ValueError('specify only one of host or hosts')
        elif (hosts is None) and (host is not None):
            self.hosts = [host]
        elif (hosts is not None) and (host is None):
            self.hosts = None
        else:
            # None here means that we should use what we find.
            self.hosts = hosts
        if not name.startswith('/') and 'parent' not in kwargs:
            vmc = kwargs['config_layout'].vmware
            kwargs['parent'] = f'/{vmc.datacenter}/datastore'
            self.parent_type = DataStoreFolder
        kwargs['name'] = name
        self.local_path = local_path
        super().__init__(**kwargs)

    async def do_create(self):

        if self.hosts is None:
            raise ValueError(f'must specify host(s) when creating datastore {self.name}')
        if len(self.hosts) != 1:
            raise NotImplementedError(
                f'support for multiple hosts is not yet implemented when creating datastore {self.name}')

        ds = self.hosts[0].mob.configManager.datastoreSystem.CreateNasDatastore(spec=self.spec)
        return
        try:
            task = self.parent.mob.MoveIntoFolder_Task([ds])
            await carthage.vmware.utils.wait_for_task(task)
        except BaseException:
            ds.DestroyDatastore()
            raise

    async def delete(self):
        task = self.mob.Destroy_Task()
        await carthage.vmware.utils.wait_for_task(task)

    #: List of ssh options to use when contacting a remote host
    ssh_opts = ('-oStrictHostKeyChecking=no', )

    @memoproperty
    def ssh_host(self):
        host, sep, path = self.local_path.partition(":")
        if sep != "":
            return host

    async def makedirs(self, dest):
        if self.ssh_host:
            host, sep, folder = self.local_path.partition(":")
            folder = Path(folder) / dest
            folder = folder.replace('"', '\\"')
            await sh.ssh(self.ssh_opts,
                         host, "mkdir", "-p",
                         f'"{folder}"',
                         _bg=True,
                         _bg_exc=False)
        else:
            os.makedirs(Path(self.local_path) / dest, exist_ok=True)

    def upload_file(self, path, dest):
        logger.info(f'Uploading {path} to {self.name}')
        params = dict(
            dsName=self.name,
            dcPath=self.config_layout.vmware.datacenter)
        verify = self.config_layout.vmware.validate_certs
        host = self.config_layout.vmware.hostname
        url = f'https://{host}:443/folder/{dest}'
        with open(path, "rb") as f:
            resp = requests.put(url, params=params,
                                data=f, auth=self.auth_tuple,
                                verify=verify)
            if resp.status_code >= 400:
                raise ValueError(f'Uploading {basename} failed: {resp.status_code}')

    async def copy_in(self, src, dest):
        '''
        Copy files into the datastore

        Copy *src* to *dest* within the datastore.  *dest* must be a relative path.
        '''
        src = Path(src)
        dest = Path(dest)
        if dest.is_absolute():
            raise ValueError("Destination cannot be absolute")
        if self.local_path and self.ssh_host:
            local_path = self.local_path
            await self.makedirs(dest.parent)
            logger.debug(f'copying {src} to {local_path}/{dest}')
            await sh.scp(self.ssh_opts, src, f'{local_path}/{dest}',
                         _bg=True, _bg_exc=False)
        elif self.local_path:
            await self.makedirs(dest.parent)
            await asyncio.get_event_loop().run_in_executor(None, shutil.copy, src, Path(self.local_path) / dest)
        else:
            await asyncio.get_event_loop().run_in_executor(None, self.upload_file, src, str(dest))

    @memoproperty
    def auth_tuple(self):
        return (self.config_layout.vmware.username, self.config_layout.vmware.password)



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

from .inventory import *

__all__ = ['VmwareFolder', 'HostFolder']


class VmwareFolder(VmwareNamedObject):

    '''
If a fully qualified name (in the form that ``searchIndex.FindByInventoryPath`` would recognize is passed in, then that is used directly as the :meth:`vmware_path`.  If a path is passed in without a leading slash, then we look up ``vmware.datacenter`` in the :class:`.ConfigLayout`  and start from the appropriate kind of folder for that data center.
    '''

    def __init_subclass__(cls, kind=NotImplemented, **kwargs):
        kwargs['kind'] = kind
        from . import datacenter
        if kind is not NotImplemented:
            cls.parent_type = (cls, datacenter.VmwareDatacenter)
        super().__init_subclass__(**kwargs)

    def __init__(self, name, *args, **kwargs):
        from . import datacenter
        if name.startswith('/') or 'parent' in kwargs:
            kwargs['name'] = name
        else:  # No initial slash, no explicit parent
            dc_name = kwargs['config_layout'].vmware.datacenter
            name = f'/{dc_name}/{self.stamp_type}/{name}'
            kwargs['name'] = name
        super().__init__(*args, **kwargs)
        if self.parent:
            self.parent_type = type(self.parent)
        else:
            if '/' in self.parent_path[1:]:
                self.parent_type = type(self)
            else:
                self.parent_type = datacenter.VmwareDatacenter

    async def do_create(self):
        if not self.parent:
            raise ValueError(f"unable to create folder '{self.name}' as no parent was specified")
        return self.parent.mob.CreateFolder(self.name)

    async def delete(self):
        task = self.mob.Destroy_Task()
        await carthage.vmware.utils.wait_for_task(task)


class HostFolder(VmwareFolder, kind='host'):
    pass



# Copyright (C)  2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations

import contextvars
import dataclasses
import traceback

_current_instantiation = contextvars.ContextVar('current_instantiation', default=None)

__all__ = []


class GetItemSet(set):

    def __getitem__(self, idx):
        self_iter = iter(self)
        result = None
        while idx >= 0:
            result = next(self_iter)
            idx -= 1
        return result


instantiation_roots = GetItemSet()

__all__ += ['instantiation_roots']


class BaseInstantiationContext:

    def __init__(self, injector):
        self.injector = injector
        self.dependencies_waiting = {}
        self.parent = None
        self._done = False

    def dependency_progress(self, key, context):
        '''Indicate that this instantiation has a dependency on *key*, currently in progress, with state tracked by *context*.'''
        self.dependencies_waiting.setdefault(key, context)

    def dependency_final(self, key, context):
        try:
            del self.dependencies_waiting[key]
        except KeyError:
            pass

    def __enter__(self):
        self.parent = _current_instantiation.get()
        if not self.parent:
            instantiation_roots.add(self)
        self.reset_token = _current_instantiation.set(self)
        return self

    def __exit__(self, *args):
        _current_instantiation.reset(self.reset_token)
        return False

    def done(self):
        '''Indicate that the instantiation has completed'''
        assert not self._done
        self._done = True
        if not self.parent:
            instantiation_roots.remove(self)

    def __str__(self):
        res = self.description
        if self.parent:
            res = f'{str(self.parent)} -> {res}'
        return res

    def __repr__(self):
        return f'{self.__class__.__name__}<{str(self)}>'


__all__ += ['BaseInstantiationContext']


@dataclasses.dataclass
class InjectedDependencyInspector:

    '''
    Using :func:`~carthage.dependency_injection.inject`, a dependency can be injected into a function or class.  Various APIs such as :meth:`~carthage.dependency_injection.introspection.get_dependencies_for` or :func:`get_dependencies_for` will allow introspection of these dependencies.  Those methods iterate over *INjectedDependencyInspector*s to give information on each dependency.
    '''

    #: The :class:`Injector` that provides the dependency. *None* if the dependency is unresolved.
    injector: typing.Optional[Injector]

    provider: DependencyProvider

    key: InjectionKey  # : What dependency is injected

    @property
    def is_provided(self):
        '''True if the dependency is provided; False if there is no injector in the chain providing this dependency.'''
        return self.provider is not None

    @property
    def is_final(self):
        '''True if the dependency has been fully instantiated.  If true, :meth:`get_value` will never raise *AsyncRequired*.
'''
        return self.is_provided and not self.provider.is_factory

    def all_keys(self):
        '''Iterate over all keys that provide this dependency in the given injector.
'''
        if not self.is_provided:
            yield self.key
        else:
            yield from self.provider.keys

    def get_value(self, ready=None):
        '''Returns the value provided for the dependency.  May raise *KeyError* if not provided.
'''
        from .base import InjectionKey
        key = self.key
        if ready is not None:
            key = InjectionKey(key, _ready=ready)
        return self.injector.get_instance(key)

    def get_value_no_instantiate(self):
        '''Return the current value without instantiating; I.E. possibly whatever is passed into add_provider.
        '''
        return self.provider.provider

    async def get_value_async(self, ready=None):
        '''Like :func:`get_value` but asynchronous.
        '''
        from .base import AsyncInjector, InjectionKey
        ainjector = AsyncInjector(self.injector)
        key = self.key
        if ready is not None:
            key = InjectionKey(key, _ready=ready)
        return await ainjector.get_instance_async(key)

    @property
    def instantiation_contexts(self):
        '''Return any ongoing instantiations referencing this dependency.  Typically contexts are very short lived if :meth:`is_final` is true.  However if a dependency is itself waiting for dependencies, the contexts can be used to chain those dependencies.
'''
        if not self.provider:
            return frozenset()
        return frozenset(self.provider.instantiation_contexts)

    def all_waiting_dependencies(self):
        result = {}
        for c in self.instantiation_contexts:
            result.update(c.dependencies_waiting)
        return result


    @property
    def provider_id(self):
        "If two inspectors have the same provider_id, they are guaranteed to refer to the same value.  It is possible that inspectors with different provider_id may refer to the same value.  Once is_final returns true, value_id will be stable."
        if not self.is_provided: return None
        return id(self.provider)
    
        

__all__ += ['InjectedDependencyInspector']


class InstantiationContext(BaseInstantiationContext, InjectedDependencyInspector):

    '''
    Represents the instantiation of an :class:`InjectionKey` in the scope of a :class:`Injector`.
    '''

    def __init__(self, injector: injector, triggering_injector: Injector, key: InjectionKey,
                 provider: DependencyProvider,
                 ready: bool):
        BaseInstantiationContext.__init__(self, injector)
        InjectedDependencyInspector.__init__(self, injector=injector, key=key, provider=provider)
        self.triggering_injector = triggering_injector
        self.ready = ready

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other

    def __enter__(self):
        for ctx in self.provider.instantiation_contexts:
            # We only care about the first one
            assert not self.dependencies_waiting
            self.dependencies_waiting = ctx.dependencies_waiting
            break
        self.provider.instantiation_contexts.add(self)
        return super().__enter__()

    @property
    def description(self):
        desc = f'Instantiating {self.key} using {self.injector}'
        if self.injector is not self.triggering_injector:
            desc += f' for {self.triggering_injector}'
        return desc

    def done(self):
        super().done()
        self.provider.instantiation_contexts.remove(self)

    def progress(self):
        for ctx in self.provider.instantiation_contexts:
            parent = ctx.parent
            if not parent:
                continue
            parent.dependency_progress(self.key, self)
        self.injector.emit_event(self.key, "dependency_progress",
                                 self,
                                 adl_keys=self.provider.keys | {base.InjectionKey(base.Injector)})

    def final(self):
        from .base import is_obj_ready
        obj_ready = is_obj_ready(self.provider.provider)
        for ctx in self.provider.instantiation_contexts:
            parent = ctx.parent
            if parent is None:
                continue
            if ctx.ready and not obj_ready:
                continue
            # We try to avoid calling dependency_final when the object
            # is not ready yet.  However since instantiations can
            # share the dependencies_waiting dict, there are cases
            # where dependency_finaly will be called anyway.  It's
            # important that when _handle_async handles a future, it
            # calls progress soon to recover from that.
            parent.dependency_final(self.key, self)
        if obj_ready or not self.ready:
            self.injector.emit_event(self.key,
                                     "dependency_final", self,
                                     adl_keys=self.provider.keys | {base.InjectionKey(base.Injector)})

    def get_dependencies(self):
        return get_dependencies_for(self.provider.provider, self.injector)


__all__ += ['InstantiationContext']


def current_instantiation():
    return _current_instantiation.get()


__all__ += ['current_instantiation']


class AsyncBecomeReadyContext(BaseInstantiationContext):

    def __init__(self, obj):
        super().__init__(obj.injector)
        self.obj = obj
        self.entered = False

    @property
    def description(self):
        return f'{self.obj}.async_become_ready()'

    def __enter__(self):
        parent = _current_instantiation.get()
        if isinstance(parent, InstantiationContext) and parent.provider.provider is self.obj:
            return parent
        self.entered = True
        return super().__enter__()

    def __exit__(self, *args):
        if self.entered:
            super().__exit__(*args)
            self.done()
        return False

    def get_dependencies(self):
        return get_dependencies_for(self.obj, self.obj.injector)


__all__ += ['AsyncBecomeReadyContext']


def get_dependencies_for(obj, injector):
    if not hasattr(obj, '_injection_dependencies'):
        return
    for injection_key in obj._injection_dependencies.values():
        try:
            dp, satisfy_against = injector._get_parent(injection_key)
            yield InjectedDependencyInspector(
                injector=satisfy_against,
                key=injection_key,
                provider=dp)
        except KeyError:
            yield InjectedDependencyInspector(
                injector=injector,
                key=injection_key,
                provider=None)


__all__ += ['get_dependencies_for']


@dataclasses.dataclass
class FailedInstantiation:

    exception: Exception
    dependency: tuple[base.InjectionKey] = dataclasses.field(default_factory=lambda: tuple())

    def __repr__(self):
        return \
            f"<Failed  Instantiation (dependency Path: {self.dependency}\n"\
            +"\n".join(traceback.format_exception(self.exception))+">"
        
def failed_instantiation(context, exception):
    injector = context.injector
    key = context.key
    dependency = tuple()
    if isinstance(exception, base.InjectionFailed):
        dependency = exception.dependency_path
        exception = exception.__context__
    failure = FailedInstantiation(exception=exception, dependency=dependency)
    all_instantiation_failures[id(injector), key] = failure
    if len(dependency) == 0:
        failed_instantiation_leaves[id(injector), key] = failure
    injector.emit_event(
        key, "dependency_instantiation_failed",
        failure,
        adl_keys={base.InjectionKey(base.Injector)})
    

__all__ += ['failed_instantiation']

all_instantiation_failures: dict[tuple, FailedInstantiation] = {}
failed_instantiation_leaves: dict[tuple, FailedInstantiation] = {}

__all__ += ['all_instantiation_failures', 'failed_instantiation_leaves']



from . import base



# Copyright (C) 2019, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import logging
import time
from dataclasses import dataclass
import carthage.ansible
import carthage.network
import os.path
from ..machine import Machine
from ..dependency_injection import *
from ..config import ConfigLayout, config_key
from ..setup_tasks import SetupTaskMixin, setup_task, SkipSetupTask
from ..utils import memoproperty
from .image import VmwareDataStore, VmdkTemplate
from .inventory import VmwareStampable
from .connection import VmwareConnection
from .folder import VmwareFolder
from . import network

logger = logging.getLogger('carthage.vmware')


class VmwarePrincipal(object):
    def __init__(self, principal):
        self.principal = principal


class VmwareUser(VmwarePrincipal):
    @property
    def group(self): return False


class VmwareGroup(VmwarePrincipal):
    @property
    def group(self): return True


class VmwareAuthorizationRole(VmwareStampable, kind='authorization_role'):

    def __init__(self, name, privIds=None, *args, **kwargs):
        self.name = name
        self.privIds = privIds
        VmwareStampable.__init__(self, *args, **kwargs)

    @memoproperty
    def stamp_descriptor(self):
        return self.name

    def _find_role(self):
        am = self.connection.content.authorizationManager
        for role in am.roleList:
            if role.name == self.name:
                self.mob = role
                return role

    @setup_task("create role")
    async def create(self):
        am = self.connection.content.authorizationManager
        role = self._find_role()
        if role is None:
            if self.privIds is None:
                raise KeyError
            am.AddAuthorizationRole(name=self.name, privIds=self.privIds)
        else:
            if self.privIds is not None:
                am.UpdateAuthorizationRole(roleId=role.roleId, newName=self.name, privIds=self.privIds)

    @create.invalidator()
    async def create(self, **kwargs):

        def lists_equal(a, b):
            return list_is_subset(a, b) and list_is_subset(b, a)

        def list_is_subset(a, b):
            for x in a:
                if x not in b:
                    return False
            return True

        role = self._find_role()
        if role is not None and self.privIds is not None:
            if not lists_equal(role.privilege, self.privIds):
                return None
        return role


@dataclass
class VmwarePermission():
    principal: VmwarePrincipal
    role: VmwareAuthorizationRole
    propagate: bool


@inject(ainjector=AsyncInjector, config=ConfigLayout)
async def create_roles(*, ainjector, config):

    cr = create_roles

    await ainjector(VmwareAuthorizationRole,
                    name='Carthage Content Creation',
                    privIds=cr.view + cr.create + cr.vm + cr.datastore + cr.globals)

    await ainjector(VmwareAuthorizationRole,
                    name='Carthage Visibility',
                    privIds=cr.view + cr.globals)

    await ainjector(VmwareAuthorizationRole,
                    name='Carthage Maintenance',
                    privIds=cr.view + cr.maintain + cr.vm + cr.datastore + cr.globals)

    await ainjector(VmwareAuthorizationRole,
                    name='Carthage Content Owner',
                    privIds=cr.view + cr.create + cr.manage +
                    cr.maintain + cr.delete + cr.vm +
                    cr.packer + cr.datastore + cr.globals)

create_roles.globals = [
    'Global.SetCustomField'
]
create_roles.view = [
    'System.View',
    'System.Anonymous',
    'System.Read'
]
create_roles.create = [
    'Datastore.Move',
    'Host.Config.Storage',
    'Host.Inventory.CreateCluster',
    'Host.Inventory.EditCluster',
    'Folder.Create',
    'DVPortgroup.Create'
]
create_roles.manage = [
    'Global.SetCustomField',
    'Authorization.ModifyPermissions',
    'VirtualMachine.Inventory.Move',
    'Host.Inventory.MoveHost',
    'Host.Inventory.EditCluster',
    'Host.Config.Maintenance'
]
create_roles.maintain = [
    'Resource.HotMigrate',
    'Resource.ColdMigrate'
]
create_roles.delete = [
    'Host.Inventory.DeleteCluster',
    'Folder.Delete',
    'DVPortgroup.Delete'
]
# Based on https://github.com/jetbrains-infra/packer-builder-vsphere/issues/97
create_roles.packer = [
    # Create VM:
    'VirtualMachine.Inventory.Create',
    'VirtualMachine.Inventory.Delete',
    'VirtualMachine.Config.AddNewDisk',
    # Customize hardware:
    'VirtualMachine.Config.CPUCount',
    'VirtualMachine.Config.Resource',
    'VirtualMachine.Config.Memory',
    'VirtualMachine.Config.Settings',
    'VirtualMachine.Config.Annotation',
    # configuration_parameters
    'VirtualMachine.Config.AdvancedConfig',
    # Boot
    'VirtualMachine.Config.Settings',
    'VirtualMachine.Interact.PowerOn',
    'VirtualMachine.Interact.ConsoleInteract',
    'VirtualMachine.Interact.PowerOff',
    # CD-ROM
    'VirtualMachine.Config.AddRemoveDevice',
    'VirtualMachine.Interact.SetCDMedia',
    'VirtualMachine.Interact.DeviceConnection',
    # Upload floppy image
    'VirtualMachine.Config.AddRemoveDevice',
    'VirtualMachine.Interact.SetFloppyMedia',
    # Snapshot:
    'VirtualMachine.State.CreateSnapshot',
    # Template:
    'VirtualMachine.Provisioning.MarkAsTemplate'
]
create_roles.vm = [
    'VirtualMachine.Config.Rename',
    'VirtualMachine.Config.Rename',
    'VirtualMachine.Config.Annotation',
    'VirtualMachine.Config.Annotation',
    'VirtualMachine.Config.AddNewDisk',
    'VirtualMachine.Config.RemoveDisk',
    'VirtualMachine.Config.RawDevice',
    'VirtualMachine.Config.HostUSBDevice',
    'VirtualMachine.Config.CPUCount',
    'VirtualMachine.Config.Memory',
    'VirtualMachine.Config.AddRemoveDevice',
    'VirtualMachine.Config.EditDevice',
    'VirtualMachine.Config.Settings',
    'VirtualMachine.Config.Resource',
    'VirtualMachine.Config.UpgradeVirtualHardware',
    'VirtualMachine.Config.ResetGuestInfo',
    'VirtualMachine.Config.ToggleForkParent',
    'VirtualMachine.Config.AdvancedConfig',
    'VirtualMachine.Config.DiskLease',
    'VirtualMachine.Config.SwapPlacement',
    'VirtualMachine.Config.DiskExtend',
    'VirtualMachine.Config.ChangeTracking',
    'VirtualMachine.Config.QueryUnownedFiles',
    'VirtualMachine.Config.ReloadFromPath',
    'VirtualMachine.Config.QueryFTCompatibility',
    'VirtualMachine.Config.MksControl',
    'VirtualMachine.Config.ManagedBy',
    'VirtualMachine.Inventory.Create',
    'VirtualMachine.Inventory.Delete',
    'VirtualMachine.Interact.PutUsbScanCodes',
    'VirtualMachine.Config.AddExistingDisk',
    'VirtualMachine.Config.AddRemoveDevice',
    'VirtualMachine.Config.AddNewDisk',
    'VirtualMachine.Interact.PowerOn',
    'VirtualMachine.Interact.PowerOff',
    'VirtualMachine.Inventory.CreateFromExisting',
    'VirtualMachine.Provisioning.CreateTemplateFromVM',
    'VirtualMachine.Provisioning.DeployTemplate',
    'VirtualMachine.Provisioning.CloneTemplate',
    'VirtualMachine.Provisioning.Clone',
    'VirtualMachine.Provisioning.MarkAsVM',
    'Resource.AssignVMToPool',
    'Network.Assign',
    'VirtualMachine.State.CreateSnapshot',
    'VirtualMachine.State.RevertToSnapshot',
    'VirtualMachine.Interact.AnswerQuestion',
    'VirtualMachine.Interact.ConsoleInteract',
    'VirtualMachine.Interact.DeviceConnection',
    'VirtualMachine.Interact.Reset',
    'VirtualMachine.Interact.SetCDMedia',
    'VirtualMachine.Interact.SetFloppyMedia',
    'VirtualMachine.Interact.ToolsInstall',
    'VApp.Import'
]
create_roles.datastore = [
    'Datastore.AllocateSpace',
    'Datastore.Browse',
    'Datastore.DeleteFile',
    'Datastore.FileManagement',
    'Datastore.UpdateVirtualMachineFiles'
]
create_roles.datastore_view = [
    'Datastore.Browse'
]



# Copyright (C) 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.
import dataclasses
import os
import shutil
import pytest
from pathlib import Path
import carthage
from carthage.pytest import *
from carthage import *
from carthage.kvstore import *
from carthage.modeling import *

class TestAssignments(HashedRangeAssignments):

    __test__ = False

    def __init__(self, objs, **kwargs):
        super().__init__(domain="test_domain", **kwargs)
        self.objs = objs

    def hash_key(self, key, obj):
        if hasattr(obj, 'hash'):
            assert obj.low <= obj.hash <= obj.high
            return obj.low, obj.hash, obj.high
        return super().hash_key(key, obj)

    def do_assignments(self):
        self.new_assignments()
        for o in self.objs:
            self._assign(o.key, o)

    def record_assignment(self, key, obj, assignment):
        obj.assignment = int(assignment)

    def valid_key(self, key):
        for o in self.objs:
            if o.key == key: return True
        return False

    def find_bounds(self, obj):
        return obj.low, obj.high

    def check_consistency(self):
        assigned = set()
        for o in self.objs:
            assert o.assignment is not None
            assert o.assignment not in assigned
            assigned.add(o.assignment)
            assert o.low <= o.assignment <= o.high
            

key_counter = 0
def next_key():
    global key_counter
    key_counter +=1
    return f'k{key_counter}'


@dataclasses.dataclass
class AssignedObj:

    low: int
    high: int
    assignment: int = None
    key: str = dataclasses.field(default_factory=next_key)

state_dir = Path(__file__).parent.joinpath("test_state")


@pytest.fixture()
def ainjector(ainjector):
    ainjector = ainjector.claim("test_kvstore.py")
    config = ainjector.injector(carthage.ConfigLayout)
    config.state_dir = state_dir
    os.makedirs(state_dir, exist_ok=True)
    ainjector.add_provider(KvStore)
    yield ainjector
    shutil.rmtree(state_dir, ignore_errors=True)

@async_test
async def test_do_assign(ainjector):
    o1 = AssignedObj(0,4)
    o2 = AssignedObj(2,3)
    objs = [o1, o2]
    assignments = await ainjector(TestAssignments, objs)
    assignments.do_assignments()
    assignments.check_consistency()


@async_test
async def test_exhaustion(ainjector):
    o1 = AssignedObj(0,1)
    o2 = AssignedObj(0,1)
    o3 = AssignedObj(0,1)
    objs = [o1, o2, o3]
    assignments = await ainjector(TestAssignments, objs)
    with pytest.raises(AssignmentsExhausted): assignments.do_assignments()

@async_test
async def test_reallocate_when_needed(ainjector):
    o1 = AssignedObj(0,1)
    o1.hash = 0
    o2 = AssignedObj(0,1)
    o2.hash = 1
    o3 = AssignedObj(0,1)
    objs = [o1,o2]
    assignments = await ainjector(TestAssignments, objs)
    assignments.do_assignments()
    assert o1.assignment == 0
    assert o2.assignment == 1
    assignments.check_consistency()
    assignments.objs = [o2,o3]
    with pytest.raises(AssignmentsExhausted):
        # exhausted because without confirming all keys are known, we
        # don't know that o1 has been removed
        assignments.do_assignments()
    assignments.enable_key_validation()
    assignments.do_assignments()
    assignments.check_consistency()
    
@async_test
async def test_reallocation_minimized(ainjector):
    "If an object is removed for a temporary time it should not lose its assignment if resources are available"
    o1 = AssignedObj(1,4)
    o2 = AssignedObj(1,4)
    o3 = AssignedObj(1,4)
    o1.hash = 1
    o2.hash = 1
    o3.hash = 1
    #Everyone wants the same assignment
    objs = [o1,o2]
    assignments = await ainjector(TestAssignments, objs)
    assignments.enable_key_validation()
    assignments.do_assignments()
    assignments.check_consistency()
    assert o1.assignment == 1
    assignments.objs = [o3, o2]
    assignments.do_assignments()
    # and should still be consistent with o1 added
    assignments.objs.append(o1)
    assignments.check_consistency()
    # But we get different behavior if we prefer reallocation
    assignments.prefer_reallocate = True
    o4 = AssignedObj(1,4)
    o4.hash = 1
    assignments.objs = [o4,o2, o3]
    assignments.do_assignments()
    assert o4.assignment == 1
    assignments.check_consistency()

@async_test
async def test_dump_load(ainjector):
    o1 = AssignedObj(1,5)
    o2 = AssignedObj(2,6)
    o3 = AssignedObj(1,6)
    o4 = AssignedObj(3,5)
    objs = [o1, o2, o3, o4]
    assignments = await ainjector(TestAssignments, objs)
    assignments.do_assignments()
    kvstore = ainjector.get_instance(KvStore)
    kvstore.dump(state_dir/'dump.yml', lambda d, k,v: True)
    correct_assignments = {o.key:o.assignment for o in objs}
    with kvstore.environment.begin(write=True) as txn, txn.cursor() as csr:
        csr.first()
        while csr.delete(): pass
    kvstore.load(state_dir/'dump.yml')
    assignments2 = await ainjector(TestAssignments, objs)
    for o in objs: o.hash = 1 # so that without db do_assignments produces different results.
    assignments2.do_assignments()
    for o in objs:
        assert o.assignment == correct_assignments[o.key]
        
    
class layout(CarthageLayout):
    class config(NetworkConfigModel):
        add('eth0', mac=None, net=injector_access('pool_network'),
            )

    @provides('pool_network')
    class pool_network(NetworkModel):
        v4_config = V4Config(
            network='192.168.1.0/24',
            dhcp=True,
            pool=(
                '192.168.1.10', '192.168.1.90'),
        )

    class a(MachineModel): pass
    class b(MachineModel): pass

@async_test
async def test_network_pool(ainjector):
    ainjector.add_provider(layout)
    l = await ainjector.get_instance_async(layout)
    kvstore = ainjector.get_instance(KvStore)
    l.pool_network.assign_addresses()
    



# Copyright (C) 2018, 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import abc
import asyncio
import contextlib
import os
import os.path
import shlex
import tempfile
import typing
from pathlib import Path

from .dependency_injection import *
from .config import ConfigLayout
from .ssh import SshKey, SshAgent, RsyncPath
from .utils import memoproperty
from . import sh
import carthage.ssh
from .setup_tasks import SetupTaskMixin, setup_task, TaskWrapperBase, TaskInspector
import logging
logger = logging.getLogger("carthage")


class MachineRunning:

    async def __aenter__(self):
        if self.machine.with_running_count <= 0:
            if self.machine.running is None:
                await self.machine.is_machine_running()
            self.machine.already_running = self.machine.running
        self.machine.with_running_count += 1
        if self.machine.running:
            if self.machine._ssh_online_required and self.ssh_online:
                await self.machine.ssh_online()
            return
        try:
            await self.machine.start_machine()
            if self.ssh_online:
                await self.machine.ssh_online()
            return

        except BaseException:
            self.machine.with_running_count -= 1
            raise

    async def __aexit__(self, exc, val, tb):
        self.machine.with_running_count -= 1
        if self.machine.with_running_count <= 0 and not self.machine.already_running:
            self.machine.with_running_count = 0
            await self.machine.stop_machine()

    def __init__(self, machine, *,
                 ssh_online=None):
        self.machine = machine
        if ssh_online is None:
            ssh_online = machine.machine_running_ssh_online
        self.ssh_online = ssh_online


ssh_origin = InjectionKey('ssh-origin')
ssh_origin_vrf = InjectionKey("ssh-origin-vrf")


class SshMixin:
    '''An Item that can be sshed to.  Will look for the ssh_origin
    injection key.  If found, this should be a container.  The ssh will be
    launched from within the network namespace of that container in order
    to reach the appropriate devices.  Requires ip_address to be made
    available.  Requires an carthage.ssh.SshAgent be injectable.
    '''

    _ssh_online_required = True

    @memoproperty
    def ip_address(self):
        '''The IP address or name at which this machine should be managed.'''
        try:
            return self.model.ip_address
        except AttributeError:
            raise NotImplementedError from None

    @memoproperty
    def ssh_options(self):
        if hasattr(self.model, 'ssh_options'):
            return self.model.ssh_options
        return ('-lroot', )

    @memoproperty
    def ssh_online_retries(self):
        if hasattr(self.model, 'ssh_online_retries'):
            assert isinstance(self.model.ssh_online_retries, int), "ssh_online_retries must be `int`"
            return self.model.ssh_online_retries
        return 60

    @memoproperty
    def ssh_online_timeout(self):
        if hasattr(self.model, 'ssh_online_timeout'):
            assert isinstance(self.model.ssh_online_timeout, int), "ssh_online_timeout must be `int`"
            return self.model.ssh_online_timeout
        return 5

    @memoproperty
    def ssh(self):
        from .network import access_ssh_origin
        try:
            ssh_origin_container = self.injector.get_instance(InjectionKey(ssh_origin, _optional=True))
        except InjectionFailed:
            from .container import Container
            ssh_origin_container = self if isinstance(self, Container) else None
        ssh_key = self.injector.get_instance(InjectionKey(carthage.ssh.SshKey, _optional=True))
        if ssh_key:
            ssh_agent = ssh_key.agent
            if ssh_key.key_path:
                key_options = ("-i", ssh_key.key_path,)
            else:
                key_options = ('',)
        else:
            ssh_agent = self.injector.get_instance(carthage.ssh.SshAgent)
            key_options = tuple()
        options = self.ssh_options + ('-oUserKnownHostsFile=' +
                                      os.path.join(self.config_layout.state_dir, 'ssh_known_hosts'),)
        if ssh_origin_container is not None:
            ip_address = self.ip_address
            ssh_origin_container.done_future().add_done_callback(self.ssh_recompute)
            return self.injector(access_ssh_origin).bake(
                "/usr/bin/ssh",
                *key_options,
                *options,
                *self.config_layout.global_ssh_options.split(),
                ip_address,
                _env=ssh_agent.agent_environ)
        else:
            return sh.ssh.bake(*key_options,
                               *options,
                               *self.config_layout.global_ssh_options.split(),
                               self.ip_address,
                               _env=ssh_agent.agent_environ)

    def rsync(self, *args):
        '''
        Call rsync with given arguments.
        An argument may be a :class:`.RsyncPath` generated by :meth:`rsync_path`.  Such a path encapsulates a host name and a path.  When *rsync* is called, Carthage
        finds the appropriate ssh_origin to select the right namespace for rsync.

        Typical usage::

            await machine.rsync("file",
                rsync_path("/etc/script")
            #Copy file to /etc/script on machine
        '''
        ssh_key = self.injector.get_instance(SshKey)
        return ssh_key.rsync(*args)

    def rsync_path(self, p):
        '''
A marker in a call to :meth:`rsync` indicating that *p* should be copied to or from *self*.  Interacts with the Carthage rsync machinery to select the right network namespace.
        '''
        return RsyncPath(self, p)

    #: The command run remotely by :meth:`ssh_online`
    ssh_online_command = 'date'

    async def ssh_online(self):
        logger.debug(f'Waiting for {self.name} to be ssh_online')
        online = False
        last_error = None
        await self.ainjector.get_instance_async(InjectionKey(carthage.ssh.SshKey, _optional=True)) #Instantiate in case it is async
        await self.ainjector.get_instance_async(carthage.ssh.SshAgent)
        for i in range(self.ssh_online_retries):
            try:
                await self.ssh(self.ssh_online_command,
                               _bg=True, _bg_exc=False,
                               _timeout=self.ssh_online_timeout)
            except (sh.TimeoutException, sh.ErrorReturnCode) as e:
                last_error = e
                await asyncio.sleep(1)
                continue
            online = True
            last_error = None
            self._ssh_online_required = False
            logger.debug(f'{self.name} is ssh_online')
            break
        if not online:
            raise TimeoutError("{} not online".format(self.ip_address)) from last_error

    def ssh_recompute(self, *args):
        try:
            del self.__dict__['ssh']
        except KeyError:
            pass
        self._ssh_online_required = True

    @classmethod
    def clear_ssh_known_hosts(cls, config_layout):
        try:
            os.unlink(
                os.path.join(config_layout.state_dir, "ssh_known_hosts"))
        except FileNotFoundError:
            pass

    def ssh_rekeyed(self):
        "Indicate that this host has been rekeyed"
        try:
            self.ip_address
        except NotImplementedError:
            return
        try:
            sh.ssh_keygen(
                "-R", self.ip_address,
                f=os.path.join(self.config_layout.state_dir, "ssh_known_hosts"))
        except sh.ErrorReturnCode:
            pass

class ResolvableModel(Injectable):

    '''
    In a :class:`carthage.modeling.CarthageLayout` all the models are instantiated as part of bringing the layout to ready.  This permits models to contribute to a shared understanding of networking and other global aspects of the layout.

    This class represents  a model that can be instantiated and *resolved*.  A layout will instantiate all :class:`ResolvableModel` objects in the scope of its injector that have a name constraint on their :class:`InjectionKey`.

    This class is defined here rather than in the modeling layer so that :class:`AbstractMachineModel` does not need to depend on the modeling layer.

    Subclasses of this model will typically need to override default_class_injection_key and probably supplimentary_injection_keys.
    
    '''

    async def resolve_model(self, force):
        pass

#: If this key is provided in the injector context of a :class:`NetworkedModel`, then that model assumes it is in the namespace of the :class:`NetworkedModel` or :class:`Machine` providing this key.  Rather than resolving the network config, the *network_links* property of the object providing this key is reused.
network_namespace_key = InjectionKey('carthage.machine.network_namespace', _ready=False)

class NetworkedModel(ResolvableModel):

    '''Represents something like a :class:`AbstractMachineModel` or a :class:`carthage.podman.PodmanPod` that generates a set of network_links from a :class:`~NetworkConfig`.

    When :meth:`resolve_networking` is called, if *self.injector* provides :ref:`network_namespace_key`,  then the network_links are reused from the object providing that dependency.  Typical usage is for a :class:`~carthage.oci.OciPod` or similar network namespace in which a :class:`AbstractMachineModel` will be run to provide *network_namespace_key*.
    

    '''

    network_links: typing.Mapping[str, carthage.network.NetworkLink]

    async def resolve_networking(self, force: bool = False):
        '''
            Adds all :class:`carthage.network.NetworkLink` objects specified in the :class:`carthage.network.NetworkConfig`  to the network_links property.

        :param force: if True, resolve the network config even if it has already been resolved once.

        '''
        from carthage.network import NetworkConfig
        if not force and self.network_links:
            return
        try:
            if hasattr(self, 'ainjector'):
                ainjector = self.ainjector
            else:
                ainjector = self.injector(AsyncInjector)
            network_namespace = await ainjector.get_instance_async(InjectionKey(network_namespace_key, _optional=True, _ready=False))
            if network_namespace and network_namespace is not self:
                await network_namespace.resolve_networking(force=force)
                self.network_links = network_namespace.network_links
                return
            network_config = await ainjector.get_instance_async(NetworkConfig)
        except KeyError:
            return
        if network_config is None:
            return
        result = await ainjector(network_config.resolve, self)

    async def resolve_model(self, force:bool=False):
        await self.resolve_networking(force=force)
        return await super().resolve_model(force=force)
    

class AbstractMachineModel(NetworkedModel):

    '''
    Represents properties of a machine that do not involve interacting with an implementation of that machine.  All the *AbstractMachineModels* in a layout can be instantiated to reason about things like network connections, configuration, and what machines will be built without instantiating any of the machines.  Typically if a :class:`Machine` has a model, the model will be made available either by setting the *model* property on the machine, or by providing a dependency for :class:`AbstractModel` in the injector in which the machine is instantiated.

    The most common concrete implementation of a machine model is :class:`carthage.modeling.MachineModel`.

    '''
    
    name: str

    #: If True, :meth:`Machine.start_dependencies()` will stop collecting dependencies at the injector of this model.  In the normal situation where the :class:`Machine` is instantiated within the model's dependency context, what this means is that  only system dependencies declared on the model will be started.  This may also be an :class:`InjectionKey`, an :class:`Injector`, or an :class:`Injectable`.  Se the documentation of :meth:`Machine.start_dependencies()`.
    override_dependencies: typing.Union[bool, Injector, Injectable, InjectionKey] = False

    @classmethod
    def default_class_injection_key(self):
        if hasattr(self, 'name'):
            return InjectionKey(AbstractMachineModel, host=self.name)
        else:
            return super().default_class_injection_key()

    @classmethod
    def supplementary_injection_keys(cls, k):
        name = None
        if 'host' in k.constraints:
            name =k.constraints['host']
        elif 'name' in k.constraints:
            name = k.constraints['name']
        if name:
            yield InjectionKey(ResolvableModel, name=name)
        yield from super().supplementary_injection_keys(k)
        


@inject_autokwargs(config_layout=ConfigLayout)
class Machine(AsyncInjectable, SshMixin):

    '''
    Represents a machine that can be interacted with.

    This is an abstract class representing the interface to machines independent of the technology being used to control them.  This class can be used for:

    * :class:`Locally run containers <carthage.container.Container>`

    * :class:`Local VMs <carthage.vm.Vm>` using KVM

    * Vmware :class:`~carthage.vmware.vm.Vm`
    * Containers or VMs run on other infrastructure.

    The main capabilities of this interface are to be able to start and stop machines, know their IP address, and connect via ssh.

    '''

    model: typing.Optional[AbstractMachineModel]
    name: str
    network_links: typing.Mapping[str, carthage.network.NetworkLink]

    #: Should machine_running default to calling ssh_online
    machine_running_ssh_online: bool = True

    def __init__(self, name=None, **kwargs):
        super().__init__(**kwargs)
        if name is not None:
            self.name = name
        else:
            if not hasattr(self, 'name'):
                raise TypeError(f'name must be supplied to constructor or set in the class')
        self.with_running_count = 0
        self.already_running = False
        self.sshfs_count = 0
        self.sshfs_lock = asyncio.Lock()
        self.injector.add_provider(InjectionKey(Machine), self)
        if not hasattr(self, 'model'):
            self.model = None
        self.running = None

    def machine_running(self, **kwargs):
        '''Returns a asynchronous context manager; within the context manager, the machine is expected to be running unless :meth:`stop_machine` is explicitly called.
'''
        return MachineRunning(self, **kwargs)

    @property
    def full_name(self):
        return self.config_layout.container_prefix + self.name

    @memoproperty
    def network_links(self):
        if self.model:
            return self.model.network_links
        return {}

    #: If true, use self.filesystem_access for rsync, otherwise use ssh.
    rsync_uses_filesystem_access = False

    async def start_dependencies(self):
        '''Interface point that should be called by :meth:`start_machine` to start any dependent machines such as routers needed by this machine.

         Default behavior is provided for machines with :class:`AbstractMachineModels` attached to thennn *model* property.  In this case,  any providers of ``InjectionKey(SystemDependency)`` with a name constraint are collected.  These objects are called  with an AsyncInjector.  For example :class:`carthage.system_dependency.MachineDependency` will start some other machine and optionally wait for it to become online.

        In typical usage, the *Machine* is contained in the injector
        context of its model.  Dependencies for the current machine
        may be added directly to its model's injector.  Dependencies
        shared among a group of machines may be added to injector
        contexts that contain the model.  For example::

            network.injector.add_provider(MachineDependency("router.network"))
            # But the fileserver also needs the domain controller
            network.fileserver_model.injector.add_provider(MachineDependency("domain-controller.network"))

        But in such a situation, the router itself should not depend on itself.  Two approaches are possible.  The first is to mask out the dependency in the router's model::

            ignore_system_dependency(network.router_model.injector, MachineDependency("router.network"))

        Another mechanism is available: the *override_dependencies* property of :class:`AbstractMachineModel`.  This property controls how far up the injector chain to look for dependencies:

        true
            Only consider dependencies directly defined on the model or the Machine.  Does not work correctly if the *self.injector* is not in the injector context of *self.model.injector*.

        an :class:`~Injector`
            Filter out dependencies declared between the parent of the model injector and the provided injector, inclusive.  So, consider a machine in a network in an enclave.  If the enclave's injector is provided, then injectors declared on the enclave and network will be ignored, but dependencies declared at a level larger than the enclave will still be started.

        an :class:`~Injectable`
            Filter out dependencies between the machine model and the injector contained in the Injectable inclusive.

        an :class:`~InjectionKey`
            Instantiate the key, and assume it is an Injectable.  Treat as that injectable

        For finer grain control, implementations can override this method.

        '''
        from carthage.system_dependency import MachineDependency, SystemDependency

        def callback(d):
            def cb(future):
                try:
                    future.result()
                except BaseException:
                    logger.exception(f"Error resolving {d}:")
            return cb

        def filter_dependencies(k):
            if 'name' not in k.constraints:
                return False
            if k in filter_keys:
                return False
            return True
        if not hasattr(self, 'model'):
            return
        logger.debug(f'Starting dependencies for {self.name}')
        stop_at = None
        model = self.model
        override_dependencies = model and model.override_dependencies
        if override_dependencies is True:
            stop_at = model.injector
            filter_keys = tuple()
        elif not override_dependencies:
            filter_keys = tuple()
        else:
            if isinstance(override_dependencies, InjectionKey):
                override_dependencies = await self.ainjector.get_instance_async(InjectionKey(override_dependencies, _ready=False))
            if isinstance(override_dependencies, Injectable):
                override_dependencies = override_dependencies.injector
            if not isinstance(override_dependencies, Injector):
                raise TypeError(
                    "override_dependencies must be boolean, an InjectionKey resolving to an INjectable, an Injectable, or an injector")
            filter_keys = model.injector.parent_injector.filter(
                SystemDependency, ['name'],
                stop_at=override_dependencies)
            logger.debug("Ignoring dependencies: %s", " ".join([k.name for k in filter_keys]))

        results = await self.ainjector.filter_instantiate_async(SystemDependency, filter_dependencies, ready=True, stop_at=stop_at)
        futures = []
        for k, d in results:
            future = asyncio.ensure_future(d(self.ainjector))
            future.add_done_callback(callback(d))
            futures.append(future)
        await asyncio.gather(*futures)

    def setup_task_event_keys(self):
        return self.supplementary_injection_keys(InjectionKey(Machine, host=self.name))
    
    async def start_machine(self):

        '''
        Must be overridden.  Start the machine.
        '''
        self.injector.emit_event(InjectionKey(Machine),
                                 "start_machine", self,
                                 adl_keys={InjectionKey(Machine, host=self.name)} |
                                 set(self.supplementary_injection_keys(InjectionKey(Machine, host=self.name))))

    async def stop_machine(self):
        ''' Must be overridden; stop the machine.
        '''
        self.injector.emit_event(InjectionKey(Machine),
                                 "stop_machine", self,
                                 adl_keys={InjectionKey(Machine, host=self.name)} |
                                 set(self.supplementary_injection_keys(InjectionKey(Machine, host=self.name))))
        self._ssh_online_required = True

    async def is_machine_running(self):
        '''
:return: Whether the machine is running

        Probe whether the machine is running and set self.running appropriately.  It is most important that running be set accurately when :meth:`start_machine` and :meth:`stop_machine` can start or stop the machine.  For :class:`BareMetalMachine` it is reasonable to assume that the machine is running.  This interface point should not call :meth:`ssh_online` or confirm the machine is on the network.
        '''
        raise NotImplementedError

    def __repr__(self):
        res = f"<{self.__class__.__name__} name:{self.name} "
        try:
            res += f"ip_address:{self.ip_address}"
        except Exception:
            pass
        res += ">"
        return res

    async def resolve_networking(self, force: bool = False):
        '''
        Adds all :class:`carthage.network.NetworkLink` objects specified in the :class:`carthage.network.NetworkConfig`  to the network_links property.

        :param force: if True, resolve the network config even if it has already been resolved once.

        '''
        from carthage.network import NetworkConfig
        if not force and self.network_links:
            return
        try:
            network_config = await self.ainjector.get_instance_async(NetworkConfig)
        except KeyError:
            return
        if network_config is None:
            return
        result = await self.ainjector(network_config.resolve, self.model or self)

    async def apply_customization(self, cust_class, method='apply', **kwargs):
        '''
        Apply a :class:`BaseCustomization` to this machine..

        :parameter stamp: A distinguisher for stamps created by the customization.  The stamp will include this value as well as the stamp from the :func:`setup_task`.

        '''
        customization = await self.ainjector(cust_class, apply_to=self, **kwargs)
        meth = getattr(customization, method)
        return await meth()

    def _apply_to_filesystem_customization(self, customization):
        '''
        Adapts the customization to this type of machine.  Overridden in machines that can customize a filesystem without booting.
        '''
        customization.customization_context = customization._machine_context()
        customization.run_command = self.ssh_run_command

    def ssh_run_command(self,
                        *args,
                        _bg=True,
                        _bg_exc=False):
        '''
            Ssh has really bad quoting; it effectively  removes one level of quoting from the input.
This handles quoting and  makes sure each argument is a separate argument on the eventual shell;
it works like :meth:`carthage.container.Container.container_command` and is used to give a consistent interface by :meth:`FilesystemCustomization.run_command`.
'''
        return self.ssh(
            shlex.join(args),
            _bg=_bg, _bg_exc=_bg_exc)

    async def sshfs_process_factory(self):
        return sh.sshfs(
            '-o' 'ssh_command=' + " ".join(
                str(self.ssh).split()[:-1]),
            f'{self.ip_address}:/',
            self.sshfs_path,
            '-f',
            _bg=True,
            _bg_exc=False)

    @contextlib.asynccontextmanager
    async def filesystem_access(self):
        '''
        An asynchronous context manager that makes the filesystem of the *Machine* available on a local path.

        :returns: Path at which the filesystem can be accessed while in the context.

        '''
        async with self.machine_running(ssh_online=True):
            self.sshfs_count += 1
            try:
                # Argument for correctness of locking.  The goal of
                # sshfs_lock is to make sure that two callers are not both
                # trying to spin up sshfs at the same time.  The lock is
                # never held when sshfs_count is < 1, so it will not block
                # when the coroutine that actually starts sshfs acquires
                # the lock.  Therefore the startup can actually proceed.
                # It would be equally correct to grab the lock before
                # incrementing sshfs_count, but more difficult to
                # implement because the lock must be released by time of
                # yield so other callers can concurrently access the filesystem.
                async with self.sshfs_lock:
                    if self.sshfs_count == 1:
                        self.sshfs_path = tempfile.mkdtemp(
                            dir=self.config_layout.state_dir, prefix=self.name, suffix="sshfs")
                        self.sshfs_process = await self.sshfs_process_factory()
                        for x in range(5):
                            alive, *rest = self.sshfs_process.process.is_alive()
                            if not alive:
                                await self.sshfs_process
                                raise RuntimeError  # I'd expect that to have happened from an sh exit error already
                            if os.path.exists(os.path.join(
                                    self.sshfs_path, "run")):
                                break
                            await asyncio.sleep(0.4)
                        else:
                            raise TimeoutError("sshfs failed to mount")
                yield Path(self.sshfs_path)
            finally:
                self.sshfs_count -= 1
                if self.sshfs_count <= 0:
                    self.sshfs_count = 0
                    try:
                        self.sshfs_process.process.terminate()
                    except BaseException:
                        pass
                    dir = self.sshfs_path
                    self.sshfs_path = None
                    self.sshfs_process = None
                    await asyncio.sleep(0.2)
                    with contextlib.suppress(OSError):
                        os.rmdir(dir)


@inject_autokwargs(config_layout=ConfigLayout)
class BaseCustomization(SetupTaskMixin, AsyncInjectable):

    def __init__(self, apply_to: Machine,
                 stamp=None, **kwargs):
        if isinstance(apply_to, BaseCustomization):
            apply_to = apply_to.host
        self.host = apply_to
        if not getattr(self, 'description', None):
            self.description = self.__class__.__name__
        self.stamp_stem = stamp or self.__class__.__name__
        super().__init__(**kwargs)

    @classmethod
    def default_class_injection_key(cls):
        description = cls.description or cls.__name__
        return InjectionKey(cls, description=description)

    # We do not run setup_tasks on construction
    async_ready = AsyncInjectable.async_ready

    @memoproperty
    @contextlib.asynccontextmanager
    async def customization_context(self):
        '''Can be overridden; context in which customization tasks are run.
'''
        try:
            yield
        finally:
            pass

    @property
    def stamp_path(self):
        return self.host.stamp_path

    def create_stamp(self, stamp, contents):
        stamp = f'{self.stamp_stem}-{stamp}'
        return self.host.create_stamp(stamp, contents)

    def check_stamp(self, stamp):
        stamp = f'{self.stamp_stem}-{stamp}'
        return self.host.check_stamp(stamp)

    def delete_stamp(self, stamp):
        stamp = f'{self.stamp_stem}-{stamp}'
        return self.host.delete_stamp(stamp)

    def inspect_setup_tasks(self):
        return super().inspect_setup_tasks(
            stamp_stem=self.stamp_stem+'-', instance_id=id(self.host))
    
    async def last_run(self):
        '''
        :return: the most recent time any setup task on this Customization has run against the given host. Returns false if the tasks definitely need to run.
        '''
        last_run = 0.0
        for t in self.setup_tasks:
            run_task, last_run = await t.should_run_task(self, last_run, ainjector=self.ainjector)
            if run_task:
                return False  # We're a check_completed function not a should_run function
        return last_run

    async def apply(self):
        ''' Run setup tasks against host'''
        return await self.ainjector(self.run_setup_tasks, context=self.customization_context)

    def __getattr__(self, a):
        if a in ('ssh', 'ip_address', 'start_machine', 'stop_machine',
                 "filesystem_access",
                 'model',
                 'name', 'ansible_inventory_name',
                 'machine_running', 'running',
                 'name', 'full_name',
                 'apply_customization'):
            return getattr(self.host, a)
        raise AttributeError(f"'{self}' has no attribute '{a}'")

    def __repr__(self):
        return f"<{self.__class__.__name__} description:\"{self.description}\" for {self.host.name}>"

    #: A description of the customization for inclusion in task logging
    description = ""


class MachineCustomization(BaseCustomization):

    '''A customization class for running customizations on running machines.'''

    @property
    def customization_context(self):
        return self.host.machine_running()


class ContainerCustomization(BaseCustomization):

    '''A customization class for running tasks on :class:`~carthage.container.Container` instances or :class:`~carthage.image.ImageVolume` instances without actually booting the container.  May also run on objects providing :meth:`_apply_to_container_customization` like :class:`carthage.podman.PodmanContainer`.  This is valuable for tasks used in image production that want to manipulate the filesystem.
'''

    def __init__(self, apply_to, **kwargs):
        if not hasattr(apply_to, '_apply_to_container_customization'):
            raise TypeError(f'{self.__class__.__name__} can only be applied to Containers or ImageVolumes')
        super().__init__(apply_to=apply_to, **kwargs)
        apply_to._apply_to_container_customization(self)

    @memoproperty
    def path(self):
        return self.host.volume.path

    def __getattr__(self, a):
        if a in ('container_command', ):
            return getattr(self.host, a)
        else:
            return super().__getattr__(a)


class FilesystemCustomization(BaseCustomization):

    '''
    A Customization class for interacting with the filesystem either of a :class:`carthage.container.Container`, :class:`ImageVolume` or :class:`Machine`.  If possible (for containers and image volumes), do not actually boot the machine.
'''

    def __init__(self, apply_to, **kwargs):
        super().__init__(apply_to, **kwargs)
        self.host._apply_to_filesystem_customization(self)

    @contextlib.asynccontextmanager
    async def _machine_context(self):
        async with self.host.machine_running(), self.host.filesystem_access() as path:
            self.path = path
            yield
            return


class CustomizationInspectorProxy:

    def __init__(self, obj, stamp):
        self.obj = obj
        self.stamp_stem = stamp


    def check_stamp(self, s, *args):
        return self.obj.check_stamp(self.stamp_stem+'-'+s, *args)
    

    @property
    def logger_for(self):
        return self.obj.logger_for

    @property
    def stamp_path(self):
        return self.obj.stamp_path

    def __repr__(self):
        return f'CustomizationInspectorProxy({repr(self.obj)})'
    
class CustomizationWrapper(TaskWrapperBase):

    customization: typing.Type[BaseCustomization]

    def __init__(self, customization, before=None, **kwargs):
        self.customization = customization
        if before:
            kwargs['order'] = before.order - 1
        try:
            if kwargs['order'] is None:
                del kwargs['order']
        except AttributeError:
            pass
        kwargs['description'] = getattr(customization, 'description', None) or customization.__name__
        super().__init__(**kwargs)

    @memoproperty
    def stamp(self):
        raise RuntimeError('This CustomizationTask was never assigned to a SetupTaskMixin')

    async def func(self, machine):
        await machine.apply_customization(self.customization, stamp=self.stamp)

    async def should_run_task(self, obj, dependency_last_run=0.0, *, ainjector):
        res = await obj.apply_customization(self.customization, method="last_run", stamp=self.stamp)
        # unfortunately if we return a last_run and one of our
        # dependencies has run more recently, we will continue to
        # generate unneeded task runs because we never inject our
        # dependency's last run into the customization so we never
        # update our last_run time.
        if not res: return True, dependency_last_run
        if res > dependency_last_run:
            dependency_last_run = res
        return False, dependency_last_run

    def inspect(self, obj, instance_id=None):
        if instance_id is None: instance_id = id(obj)
        proxy = CustomizationInspectorProxy(obj, self.stamp)
        prev_inspector = None
        for t in self.customization.class_setup_tasks():
            prev_inspector = TaskInspector(task=t, from_obj=proxy, previous=prev_inspector)
            prev_inspector.stamp = self.stamp+'-'+prev_inspector.stamp
            prev_inspector.instance_id = instance_id
            yield prev_inspector
            

def customization_task(c: BaseCustomization, order: int = None,
                       before=None):
    '''
    :return: a setup_task for using a particular :class:`Customization` in a given :class:`Machine`.

    Usage::

        # in a machine
        add_packages = customization_task(AddOurPackagesCustomization)

    '''
    return CustomizationWrapper(customization=c,
                                order=order,
                                before=before)


@inject_autokwargs(ssh_key=InjectionKey(SshKey, _ready=True))
class BareMetalMachine(Machine, SetupTaskMixin):

    '''Represents physical hardware that Carthage cannot start or stop
    '''

    running = False

    async def start_machine(self):
        if self.running:
            return
        await self.start_dependencies()
        await super().start_machine()
        await self.ssh_online()
        self.running = True

    async def stop_machine(self):
        await super().stop_machine()
        self.running = False

    async def async_ready(self):
        await self.resolve_networking()
        await self.run_setup_tasks()
        await super().async_ready()

    async def is_machine_running(self):
        return self.running

    @memoproperty
    def stamp_path(self):
        return Path(f'{self.config_layout.state_dir}/machines/{self.name}')


def disk_config_from_model(model, default_disk_config):
    '''Return a :ref:`disk_config <disk_config>` specification from a model.  Handles *disk_sizes* and makes sure there is always an entry for the primary disk.
'''
    primary_disk_found = False
    if hasattr(model, 'disk_config'):
        for entry in model.disk_config:
            primary_disk_found = True
            yield dict(entry)  # copy the first level of the dict
    elif hasattr(model, 'disk_sizes'):
        for size in model.disk_sizes:
            primary_disk_found = True
            yield dict(size=size)
    if not primary_disk_found:
        yield from default_disk_config


__all__ = ['AbstractMachineModel',
           'ssh_origin',
           'Machine', 'MachineRunning', 'BareMetalMachine',
           'ResolvableModel', 'NetworkedModel',
           'SshMixin', 'BaseCustomization', 'ContainerCustomization',
           'FilesystemCustomization',
           'MachineCustomization']



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

from .inventory import VmwareNamedObject


class VmwareDatacenter(VmwareNamedObject, kind='datacenter'):

    def __init__(self, *args, config_layout, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = config_layout.vmware.datacenter
            kwargs['readonly'] = kwargs.get('readonly', True)
        super().__init__(*args, **kwargs, config_layout=config_layout)

    async def async_ready(self):
        from .datastore import DataStoreFolder
        from .host import HostFolder
        from .network import NetworkFolder
        from .vm import VmFolder
        self.datastore_folder = await self.ainjector(DataStoreFolder, name='datastore', parent=self)
        self.host_folder = await self.ainjector(HostFolder, name='host', parent=self)
        self.network_folder = await self.ainjector(NetworkFolder, name='network', parent=self)
        self.vm_folder = await self.ainjector(VmFolder, name='vm', parent=self)
        ret = await super().async_ready()
        return ret

    async def do_create(self):
        content = self.connection.content
        content.rootFolder.CreateDatacenter(name=self.name)

    parent_type = type(None)

    is_root = True



# Copyright (C) 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.


import asyncio as _asyncio
import sh as _sh

'''
Monkey patch the sh module to include __await__.
Also, by default for commands retrieved through this module set _bg=True and _bg_exc=False
'''
Command = _sh.Command

def __getattr__(name):
    val = getattr(_sh, name)
    if isinstance(val, Command):
        val = val.bake(_bg=True, _bg_exc=False)
    globals()[name] = val
    return val

def running_command_await(self):
    loop = _asyncio.get_event_loop()
    res =  yield from loop.run_in_executor(None, self.wait)
    return res

if not hasattr(_sh.RunningCommand, '__await__'):
    _sh.RunningCommand.__await__ = running_command_await
del running_command_await



# Copyright (C) 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os
import os.path
import hvac
from carthage import sh
from carthage.config import ConfigSchema, ConfigLayout
from carthage.dependency_injection import *
from carthage.dependency_injection import is_obj_ready
from carthage.config.types import ConfigString, ConfigPath, ConfigLookupPlugin
from carthage.setup_tasks import setup_task
from carthage.ssh import AuthorizedKeysFile, SshAgent, SshKey
from carthage.utils import memoproperty


class VaultError(RuntimeError):
    pass


class VaultConfig(ConfigSchema, prefix="vault"):

    #: The address of the vault to contact
    address: ConfigString = None

    #: Path to a CA bundle
    ca_bundle: ConfigString

    #: The path to an ssh key in vault
    ssh_key: ConfigString


vault_token_key = InjectionKey('vault.token')


@inject(
    injector=Injector,
    token=InjectionKey("vault.token", optional=True),
)
class Vault(Injectable):

    def __init__(self, injector, token=None):
        config = injector(ConfigLayout)
        self.vault_config = config.vault
        self.injector = injector
        super().__init__()
        #: The hvac client for this vault
        self.client = None
        self.setup_client(token)

    def setup_client(self, token):
        '''

        Sets up a client interface to a vault.

            :param token: A token for accessing the vault
            :type token: str or None
        '''

        self.client = hvac.Client(url=self.vault_config. address,
                                  verify=self.vault_config.ca_bundle,
                                  token=token)

    def initialize(self, output_directory, unseal=True,
                   secret_threshold=1,
                   secret_shares=1,
                   **kwargs):
        '''Initialize a vault.
Writes unseal keys and root token to the given output directory, which hopefully is ephemeral.

        The output directory will contain:

        * ``key.``n: Unseal keys starting at index 0

        * token: root token

        '''
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)
        if not os.path.isdir(output_directory):
            raise ValueError(f'{output_directory} is not a directory')
        result = self.client.sys.initialize(**kwargs,
                                            secret_shares=secret_shares,
                                            secret_threshold=secret_threshold)
        od = output_directory
        for i, k in enumerate(result['keys_base64']):
            with open(os.path.join(od, f'key.{i}'), "wt") as f:
                f.write(f'{k}')
        with open(os.path.join(od, "token"), "wt") as f:
            f.write(result['root_token'])
        if unseal:
            for i in range(secret_threshold):
                self.client.sys.submit_unseal_key(result['keys_base64'][i])
        self.setup_client(result['root_token'])
        return result

    def apply_config(self, config):
        '''
        Apply configuration such as policies or  authentication methods to vault.

        :param dict config: A configuration dictionary; see :ref:`vault:config` for details.
        '''
        _apply_config_to_vault(self.client, config)
        return


def _apply_config_to_vault(client, config):
    config = dict(config)  # copy so we can mutate
    _apply_policy(client, config.pop('policy', {}))
    _apply_auth(client, config.pop('auth', {}))
    _apply_secrets(client, config.pop('secrets', {}))
    _apply_audit(client, config.pop('audit', {}))
    for k in config:
        try:
            client.write(k, **config[k])
        except Exception as e:
            raise VaultError(f"failed to write {k}") from e


def _apply_policy(client, policy):
    for p in policy:
        try:
            client.sys.create_or_update_policy(p, policy[p])
        except Exception as e:
            raise VaultError(f"Unable to create Policy {p}")


def _apply_auth(client, auth):
    if not auth:
        return
    auth_methods = set(client.sys.list_auth_methods()['data'].keys())
    for a, info in auth.items():
        try:
            desc = info.pop('description', '')
            method_type = info.pop('type', a)
            if a + "/" in auth_methods:
                client.sys.tune_auth_method(a, **info)
            else:
                client.sys.enable_auth_method(method_type=method_type, path=a, description=desc,
                                              config=info)
        except Exception as e:
            raise VaultError(f"Unable to enable auth method at path {a}")


def _apply_secrets(client, secrets):
    if not secrets:
        return
    secrets_engines = set(client.sys.list_mounted_secrets_engines()['data'].keys())
    for s, info in secrets.items():
        try:
            desc = info.pop('description', '')
            backend_type = info.pop('type', s)
            if s + "/" in secrets_engines:
                client.sys.tune_mount_configuration(s, **info)
            else:
                client.sys.enable_secrets_engine(backend_type=backend_type, path=s, description=desc,
                                                 config=info)
        except Exception as e:
            raise VaultError(f"Unable to enable secrets engine at path {s}")


def _apply_audit(client, audit):
    if not audit:
        return
    audit_devices = set(client.sys.list_enabled_audit_devices()['data'].keys())
    for a, info in audit.items():
        try:
            device_type = info.pop('type', a)
            if a + "/" in audit_devices:
                continue
            client.sys.enable_audit_device(
                device_type=device_type, path=a, options=info)
        except Exception as e:
            raise VaultError(f"Unable to enable auth method at path {a}")


@inject(
    vault=Vault)
class VaultConfigPlugin(ConfigLookupPlugin):

    '''
    Usage in yaml::

        password: {vault:secret/password/{host}:password}

    '''

    def __init__(self, vault):
        self.vault = vault

    def __call__(self, selector):
        client = self.vault.client
        secret, sep, field = selector.partition(':')
        if field == "":
            raise SyntaxError("The vault plugin requires a field")
        result = client.read(secret)
        return result['data'][field]


@inject_autokwargs(
    vault=Vault
)
class VaultSshKey(SshKey):

    def __init__(self, **kwargs):
        if 'key_size' in kwargs:
            self._key_size = kwargs.pop('key_size')
        else:
            self._key_size = 2048
        super().__init__(**kwargs)
        config_layout = self.injector(ConfigLayout)
        if not hasattr(config_layout.vault, 'ssh_key'):
            raise AttributeError(
                "\nYou must specify\n\nvault:\n  ssh_key: path/to/key-name\n\nfor this implementation to function")
        self._vault_key_path = config_layout.vault.ssh_key

        self._pubs = None

    def add_to_agent(self, agent):
        assert is_obj_ready(self), f"{self} is not READY"
        r = self.vault.client.read(self._vault_key_path)['data']['data']
        sh.ssh_add('-', _env=agent.agent_environ, _in=r['PrivateKey'])
        del(r)

    @setup_task('gen-key')
    async def generate_key(self):
        pk = sh.openssl(
            'genpkey',
            '-algorithm=RSA',
            '-pkeyopt',
            f'rsa_keygen_bits:{self._key_size}',
            '-outform=PEM',
            _in=None,
            _bg=True,
            _bg_exc=False)
        await pk
        pk_str = pk.stdout
        pubk = sh.openssl('pkey', '-pubout', _in=pk, _bg=True, _bg_exc=False)
        await pubk
        pubk_str = pubk.stdout
        pubs = sh.ssh_keygen('-i', '-m', 'PKCS8', '-f', '/dev/stdin', _in=pubk, _bg=True, _bg_exc=False)
        await pubs
        pubs_str = pubs.stdout

        self.vault.client.write(self._vault_key_path, **dict(data=dict(PrivateKey=pk_str,
                                                                       PublicKey=pubk_str, SshPublicKey=pubs_str)))
        self._pubs = pubs_str
        del(pk, pubk, pubs)

    @generate_key.check_completed()
    def generate_key(self):
        r = self.vault.client.read(self._vault_key_path)
        if r is None:
            return False
        r = r['data']['data']
        if ('PrivateKey' and 'PublicKey' and 'SshPublicKey') not in r.keys():
            return False
        self._pubs = r['SshPublicKey'].replace('\n', '')
        return True

    @memoproperty
    def key_path(self):
        return None

    @memoproperty
    def vault_key_path(self):
        return self._vault_key_path

    @property
    def pubkey_contents(self):
        return self._pubs


@inject(injector=Injector)
def carthage_plugin(injector):
    injector.add_provider(Vault)
    VaultConfigPlugin.register(injector, "vault")



# Copyright (C) 2018, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.


from carthage.pki import PkiManager
from carthage.config import ConfigLayout
from carthage import base_injector, Injector
import os
import os.path
import pytest
import shutil
import carthage.sh


@pytest.fixture(scope='module')
def injector():
    injector = base_injector(Injector)
    cl = injector.get_instance(ConfigLayout)
    cl.state_dir = os.path.join(os.path.dirname(__file__), "test_state")
    try:
        carthage.sh.entanglement_pki('--help')
    except BaseException:
        pytest.skip('entanglement_pki not installed')
    yield injector
    shutil.rmtree(cl.state_dir)


@pytest.fixture
def pki_manager(injector):
    return injector(PkiManager)


def test_certify(pki_manager):
    creds = pki_manager.credentials('photon.cambridge')
    assert 'CERTIFICATE' in creds
    assert 'CERTIFICATE' in pki_manager.ca_cert



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import logging
import tempfile
import time
from carthage.debian import debian_container_to_vm, install_stage1_packages_task
from carthage.image import ContainerImage
from carthage.network import Network, V4Config
from carthage.container import Container, container_image
from carthage.utils import when_needed
from carthage.pytest import *
import os.path
import pytest
import posix
from pathlib import Path
from carthage import base_injector, AsyncInjector, sh, MachineCustomization, customization_task, ConfigLayout, ContainerCustomization
from carthage.dependency_injection import *
from carthage.systemd import SystemdNetworkModelMixin
from carthage.modeling import *
from carthage.ansible import *
from carthage.setup_tasks import *
import carthage.ssh

_resource_dir = Path(__file__).parent.joinpath("resources")

pytestmark = pytest.mark.no_rootless


@pytest.fixture()
def container(test_ainjector, loop):
    ainjector = test_ainjector
    container = loop.run_until_complete(ainjector(Container, name="container-1"))
    yield container
    if container.running:
        loop.run_until_complete(container.stop_container())
    container.close()


@async_test
async def test_start_container(container, loop):
    await container.start_container()


@async_test
async def test_container_running(container, loop):
    async with container.container_running(ssh_online=False):
        container.shell("/bin/ls")


class LayoutTest(ModelGroup):

    @provides("test_net")
    class test_net(NetworkModel, AsyncInjectable):
        name = "test_net"

        async def async_ready(self):
            await super().async_ready()
            from carthage.network import BridgeNetwork
            net = await self.ainjector(self.access_by, BridgeNetwork)
            await sh.ip(
                "addr",
                "add", "10.2.0.1/24", "dev", net.bridge_name)

    class net_config(NetworkConfigModel):
        add('eth0', mac=None, net=InjectionKey("test_net", _ready=True),
            v4_config=V4Config(address="10.2.0.2",
                               network="10.2.0.0/24"))

    add_provider(machine_implementation_key, dependency_quote(Container))

    class never_running_container(MachineModel):
        name = "never-running"

        class Cust(ContainerCustomization):
            install_ansible = install_stage1_packages_task(["ansible"])
            do_roles = ansible_role_task(os.path.dirname(__file__) + "/resources/test_ansible_role")

    class test_container(MachineModel, SystemdNetworkModelMixin):
        name = "test-container"

        ip_address = "10.2.0.2"

        class test_local_ansible(ContainerCustomization):

            install_ansible = install_stage1_packages_task(["ansible"])

            @setup_task("Copy in playbook")
            def copy_in_playbook(self):
                with open(Path(self.path) / "local_playbook.yml", "wt") as f:
                    f.write(_resource_dir.joinpath("local_playbook.yml").read_text())

            local_play = ansible_playbook_task("/local_playbook.yml", origin=True)

        class cust(MachineCustomization):

            do_something = ansible_playbook_task("resources/test_ansible.yml")
            do_roles = ansible_role_task(os.path.dirname(__file__) + "/resources/test_ansible_role")


@async_test
@inject(config=ConfigLayout)
async def test_ansible_and_modeling(test_ainjector, config):
    ainjector = test_ainjector
    layout = await ainjector(LayoutTest)
    layout.injector.add_provider(when_needed(AnsibleInventory, config.state_dir + "/ansible.yml"))
    ainjector.add_provider(InjectionKey("layout"), layout)  # So it is cleaned up
    await ainjector.get_instance_async(carthage.ssh.SshKey)
    await layout.generate()
    with TestTiming(300):
        await layout.test_container.machine.async_become_ready()
        await layout.never_running_container.machine.async_become_ready()


@async_test
async def test_container_to_vm(test_ainjector):
    ainjector = test_ainjector
    image = await ainjector.get_instance_async(container_image)
    with tempfile.TemporaryDirectory() as tmp:
        await ainjector(
            debian_container_to_vm,
            volume=image,
            output=os.path.join(tmp, "container_to_vm.raw"),
            size="4G")



# Copyright (C) 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio
import contextlib
import weakref
from .utils import possibly_async


class EventScope:

    '''
    Typically most objects that have event listener support never have listeners attached.  So it is desirable to separate the ability to listen for events from the data structures associated with actually doing so.  An *EventScope* is attached to a *target* when a *target* gains the first event subscription.  When an object lower in the hierarchy gains an event subscription, then :meth:`.break_at` is called to create a new *EventScope* and reparent targets lower in the tree to that new scope.
    '''

    def __init__(self, target, parent: EventScope = None):
        self.target = weakref.ref(target)
        self.listeners = {}
        self.parent = parent
        if parent:
            self.children, self.finalizers = parent.find_prune_children(target)
        else:
            self.children = {id(target): weakref.WeakSet()}
            self.finalizers = {}
        for v in self.children.values():
            for c in v:
                if c._event_scope is parent:
                    c._event_scope = self
                else:
                    c._event_scope._update_parent(parent, self)

    def break_at(self, target: EventListener):
        '''
        :param target:  the target at which we would like to have an :class:`EventScope`

        Returns an :class:`EventScope` that receives events for the given *target* and its children, but no parent objects.
        '''
        if self.target() is target:
            return self
        scope = type(self)(target, parent=self)
        target._event_scope = scope
        return scope

    def add_child(self, parent, child):
        '''Must be called for any object that has *self* as an *EventScope* and is not *self.target*.'''
        assert id(parent) in self.children
        child._event_parent_id = id(parent)
        self.children[id(parent)].add(child)
        self.children[id(child)] = weakref.WeakSet()
        fin = weakref.finalize(child, self._finalize_child, self.children, self.finalizers, id(child))
        fin.atexit = False
        self.finalizers[id(child)] = fin

    def find_prune_children(self, new_target):
        children = {}
        finalizers = {}

        def recurse(elt):
            for c in self.children[id(elt)]:
                recurse(c)
            children[id(elt)] = self.children[id(elt)]
            del self.children[id(elt)]
            try:
                del self.finalizers[id(elt)]
            except KeyError:
                pass
            fin = weakref.finalize(elt, self._finalize_child, children, finalizers, id(elt))
            fin.atexit = False
            finalizers[id(elt)] = fin
        recurse(new_target)
        self.children[new_target._event_parent_id].remove(new_target)
        return children, finalizers

    @staticmethod
    def _finalize_child(children, finalizers, id):
        try:
            del children[id]
        except BaseException:
            pass
        try:
            del finalizers[id]
        except BaseException:
            pass

    def _update_parent(self, old_parent, new_parent):
        p = self.parent
        assert p is not old_parent
        while p is not None:
            if p is new_parent:
                return
            if p.parent is old_parent:
                p.parent = new_parent
                return
            p = p.parent
        raise ValueError('old_parent is not  in the parent chain')

    def add_listener(self, k, event, callback):
        d = self.listeners.setdefault(k, {})
        d[callback] = (event, set())

    def remove_listener(self, k, callback):
        d = self.listeners[k]
        try:
            events, futures = d[callback]
            del d[callback]
            return futures
        except KeyError:
            # We prefer our message
            raise KeyError(f'{callback} not registered as a listener for {k}') from None

    def emit(self, loop, k, event, target, *args,
             adl_keys=set(),
             **kwargs):
        def gen_callback(futures):
            def callback(future):
                # ignore the result
                try:
                    future.result()
                except BaseException:
                    pass
                futures.remove(future)
            return callback
        if not isinstance(adl_keys, set):
            adl_keys = set(adl_keys)
        target_keys = {k} | adl_keys
        result_futures = []
        if self.parent:
            result_futures.append(self.parent.emit(
                loop, k, event, target, adl_keys=adl_keys,
                **kwargs))
        for ck in target_keys:
            try:
                d = self.listeners[ck]
            except KeyError:
                continue
            for callback, (events, futures) in d.items():
                if event in events:
                    future = loop.create_task(
                        possibly_async(callback(
                            key=ck, event=event, target=target, *args,
                            target_key=k, **kwargs)))
                    result_futures.append(future)
                    futures.add(future)
                    future.add_done_callback(gen_callback(futures))
        del args
        del kwargs

        if result_futures:
            return asyncio.gather(*result_futures)
        else:
            future = loop.create_future()
            future.set_result([])
            return future


class EventListener:

    '''Represents an object to which event listeners can be attached using :meth:`add_event_listener`.  Events are dispatched using :meth:`emit_event`.  Events are named by a string, and dispatched to keys, typically :class:`carthage.InjectionKey`.

'''

    def __init__(self, event_scope=None):
        super().__init__()
        if event_scope:
            self._event_scope = event_scope
            return
        er = getattr(self, '_event_scope', None)
        if isinstance(er, EventScope):
            return
        try:
            self._event_scope = self.parent._event_scope
            self._event_scope.add_child(self.parent, self)
            return
        except AttributeError:
            pass
        self._event_scope = EventScope(self)

    def add_event_listener(self, key, events, callback):
        '''
         :param key: an :class:`InjectionKey` or similar key toward which the event will be dispatched.

        :param events:  a string (or sequence of strings) indicating which event will be  dispatched.

        :param callback: A callable that will be called as::

            callback(key, event, target, *event_args, **event_kwargs)

        The *callback* may be asynchronous.
        '''
        self._event_scope.break_at(self)
        if isinstance(events, str):
            events = {events}
        events = frozenset(events)
        self._event_scope.add_listener(key, events, callback)

    def remove_event_listener(self, key, callback):
        '''
        :return: Set of futures representing pending calls to the removed callback.

        '''
        return self._event_scope.remove_listener(key, callback)

    def emit_event(self, key, event, target,
                   *args,
                   adl_keys=set(),
                   loop=None,
                   **kwargs):
        if loop is None:
            try:
                loop = self.loop
            except BaseException:
                loop = asyncio.get_event_loop()
        return self._event_scope.emit(loop, key, event, target,
                                      *args,
                                      **kwargs,
                                      adl_keys=adl_keys,
                                      scope=self)

    @contextlib.contextmanager
    def event_listener_context(self, key, events, callback):
        '''
        Within the scope of the context, *callback* is registered as a listener for the *events* directed at *key*.
        A callback may be removed prematurely if  it is registered for the same key on the same scope by multiple calls to this function or :meth:`add_event_listener`.

        :return: A set of futures representing pending  calls to the callback.

        '''
        self.add_event_listener(key, events, callback)
        try:
            ignore, futures = self._event_scope.listeners[key][callback]
            yield futures
        finally:
            self.remove_event_listener(key, callback)


__all__ = ['EventListener']



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import itertools
import asyncio
import yaml
import urllib3
import carthage
import carthage.utils

from pyVmomi import vim, vmodl

from carthage import base_injector, inject, AsyncInjector, ConfigLayout, ansible
from carthage.config import ConfigSchema
from carthage.dependency_injection import *
from carthage.ssh import SshKey
from carthage.console import *
from carthage.network import *
from carthage.vmware import *
from carthage.vmware.network import *
from carthage.vmware.folder import *
from carthage.vmware.inventory import *
from carthage.vmware.authorization import *
from carthage.vmware.datastore import *
from carthage.vmware.datacenter import VmwareDatacenter
from carthage.vmware.cluster import VmwareCluster
from carthage.vmware.utils import wait_for_task
from carthage.utils import *


def carthage_load_config(s):
    base_injector(ConfigLayout).load_yaml(s)



# Copyright (C) 2018, 2019, 2020, 2021, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import atexit
import sys

import carthage.config
import carthage.dependency_injection

__all__ = []

from .utils import memoproperty, when_needed, relative_path
__all__ += ['memoproperty', 'when_needed', 'relative_path']

from .setup_tasks import *
__all__ += carthage.setup_tasks.__all__

from .dependency_injection import *
__all__ += carthage.dependency_injection.__all__

from .config import ConfigLayout, config_key, ConfigSchema
__all__ += ['ConfigLayout', 'config_key', 'ConfigSchema']


from .kvstore import KvStore, KvConsistency,AssignmentsExhausted, persistent_seed_path

__all__ += ['persistent_seed_path']



from .network import Network, NetworkConfig, MacStore, V4Config, V4Pool

__all__ += ['Network', 'NetworkConfig', 'MacStore', 'V4Config', 'V4Pool']



from .machine import Machine, AbstractMachineModel, MachineCustomization, ContainerCustomization, FilesystemCustomization, customization_task, BareMetalMachine
import carthage.ssh  # ssh import must come after machine
from .ssh import RsyncPath
import carthage.pki
from . import ansible
from . import cloud_init
from .files import rsync_git_tree, git_tree_hash


__all__ += ['Machine', 'rsync_git_tree',
            'git_tree_hash',
            'RsyncPath',
            'AbstractMachineModel', 'MachineCustomization',
            'ContainerCustomization', 'FilesystemCustomization',
            'customization_task',
            'BareMetalMachine']

from . import image
from .image import ContainerVolume, wrap_container_customization

__all__ += ['ContainerVolume', 'wrap_container_customization']


from .system_dependency import MachineDependency, SystemDependency, disable_system_dependency

__all__ += ['MachineDependency', 'SystemDependency', 'disable_system_dependency']

import carthage.container
import carthage.local
from .local import LocalMachine, LocalMachineMixin
__all__ += ['LocalMachine', 'LocalMachineMixin']

import carthage.pki

import carthage.vm

import carthage.debian
from .debian import DebianContainerImage

__all__ += ['DebianContainerImage']

from .dns import DnsZone, PublicDnsManagement

__all__ += ['DnsZone', 'PublicDnsManagement']

from .plugins import CarthagePlugin
from . import plugins

__all__ += ['CarthagePlugin']

base_injector = carthage.dependency_injection.Injector()
base_injector.claim("base injector")
carthage.config.inject_config(base_injector)
base_injector.add_provider(ssh.SshKey)
base_injector.add_provider(ssh.AuthorizedKeysFile)
base_injector.add_provider(asyncio.get_event_loop(), close=False)
base_injector.add_provider(KvStore)
base_injector.add_provider(MacStore)
base_injector.add_provider(ansible.AnsibleConfig)
base_injector.add_provider(carthage.network.external_network)
base_injector.add_provider(carthage.network.BridgeNetwork, allow_multiple=True)

base_injector.add_provider(InjectionKey(carthage.ssh.SshAgent), carthage.ssh.ssh_agent)
base_injector.add_provider(carthage.pki.PkiManager)
base_injector(carthage.cloud_init.enable_cloud_init_plugins)


__all__ += ['base_injector']

base_injector(plugins.load_plugin_from_package, sys.modules[__name__])

@atexit.register
def __done():
    asyncio.run(dependency_injection.shutdown_injector(base_injector))
    import sys
    sys.last_traceback = None
    sys.last_value = None
    import gc
    gc.collect()



# Copyright (C) 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import functools
import typing
from carthage.dependency_injection import Injector, InjectionKey, inject_autokwargs, dependency_quote, AsyncRequired, inject
from .implementation import ModelingBase, InjectableModelType, ModelingContainer, NSFlags
from .utils import setattr_default, fixup_dynamic_name
from ..dependency_injection import InjectionKey
import carthage.machine


class ModelingDecoratorWrapper:

    subclass: type = ModelingBase
    name: str

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'{self.__class__.name}({repr(self.value)})'

    def handle(self, cls, ns, k, state):
        if not issubclass(cls, self.subclass):
            raise TypeError(f'{self.__class__.name} decorator only for subclasses of {self.cls.__name__}')
        state.value = self.value
        if isinstance(self.value, ModelingDecoratorWrapper):
            if self.value.handle(cls, ns, k, state):
                raise TypeError("A decorator that suppresses assignment must be outermost")
            else:
                self.value = state.value  # in case superclasses care

    @property
    def __provides_dependencies_for__(self): return self.value.__provides_dependencies_for__

    _injection_error = "Use @inject on the innermost value not on a decorator."
    
class injector_access(ModelingDecoratorWrapper):

    '''Usage::

        val = injector_access("foo")

    At runtime, ``model_instance.val`` will be memoized to ``model_instance.injector.get_instance("foo")``.

    '''

    # Unlike most ModelingDecorators we want to get left in the
    # namespace.  The only reason this is a ModelingDecorator is so we
    # can clear the inject_by_name flag to avoid namespace polution and
    # to minimize the probability of a circular references from
    # something like ``internet = injector_access("internet")``

    key: InjectionKey
    name: str
    target: type

    def __init__(self, key, target=None):
        super().__init__(key)
        if not isinstance(key, InjectionKey):
            if isinstance(key, InjectableModelType) and key.__provides_dependencies_for__ and target is None:
                key = key.__provides_dependencies_for__[0]
                target = key
            else:
                key = InjectionKey(key, _ready=False)
        if key.ready is None:
            key = InjectionKey(key, _ready=False)
        #: The injection key to be accessed.
        self.key = key
        self.target = target
        # Our __call__method needs an injector
        inject_autokwargs(injector=Injector)(self)

    _injection_error = None #Because injector_access does have dependencies.

    def handle(self, cls, ns, k, state):
        super().handle(cls, ns, k, state)
        state.value = self
        state.flags &= ~NSFlags.inject_by_name

    def __get__(self, inst, owner):
        if inst is None:
            if self.target:
                return self.target
            return self
        try:
            res = inst.injector.get_instance(self.key)
            setattr(inst, self.name, res)
            return res
        except KeyError as e:
            # Injectors return KeyError, but for a missing Attribute
            # we want to raise AttributeError so for example hasattr
            # will be false.
            raise AttributeError(str(e)) from e

    def __set_name__(self, owner, name):
        self.name = name

    def __call__(self, injector: Injector):
        try:
            return injector.get_instance(self.key)
        except AsyncRequired:
            @inject(injector=None,
                    injectable=InjectionKey(Injector))
            class Xref(carthage.dependency_injection.base.InjectorXrefMarker):
                target_key = self.key

            return Xref(injectable=injector)
        

    def __repr__(self):
        return f'injector_access({repr(self.key)})'


class ProvidesDecorator(ModelingDecoratorWrapper):
    subclass = InjectableModelType
    name = "provides"

    def __init__(self, value, *keys):
        super().__init__(value)
        self.keys = keys

    def handle(self, cls, ns, k, state):
        super().handle(cls, ns, k, state)
        state.extra_keys.extend(self.keys)


def provides(*keys):
    '''Indicate that the decorated value provides these InjectionKeys'''
    keys = list(map(lambda k: InjectionKey(k), keys))

    def wrapper(val):
        try:
            setattr_default(val, '__provides_dependencies_for__', [])
            val.__provides_dependencies_for__ = keys + val.__provides_dependencies_for__
            return val
        except BaseException:

            return ProvidesDecorator(val, *keys)
    return wrapper


class DynamicNameDecorator(ModelingDecoratorWrapper):

    name = "dynamic_name"

    def __init__(self, value, new_name):
        super().__init__(value)
        self.new_name = new_name

    def handle(self, cls, ns, k, state):
        super().handle(cls, ns, k, state)
        value = self.value
        if hasattr(value, '__qualname__'):
            value.__qualname__ = ns['__qualname__'] + '.' + self.new_name
            value.__name__ = self.new_name
        state.value = value
        state.new_name = self.new_name


def dynamic_name(name):
    '''A decorator to be used in a modeling type to supply a dynamic name.  Example::

        for i in range(3):
            @dynamic_name(f'i{i+1}')
    class ignored: square = i*i

    Will define threevariables i1 through i3, with values of squares (i1 = 0, i2 =1, i3 = 4).
    '''
    name = fixup_dynamic_name(name)

    def wrapper(val):
        return DynamicNameDecorator(val, name)
    return wrapper


def globally_unique_key(
    key: typing.Union[InjectionKey, typing.Callable[[object], InjectionKey]],
):
    '''Decorate a value to indicate that *key* is a globally unique
    :class:`~InjectionKey` that should provide the given value.
    Globally unique keys are not extended with additional constraints
    when propagated up through :class:`~ModelingContainers`.

    :param key:  A callback that maps the value to a key.  Alternatively, simply the :class:`~InjectionKey` to use.

    '''
    def wrapper(val):
        nonlocal key
        if callable(key):
            key = key(val)
        elif isinstance(key, str):
            key = InjectionKey(key)
        setattr_default(val, '__provides_dependencies_for__', [])
        if key not in val.__provides_dependencies_for__:
            val.__provides_dependencies_for__.append(InjectionKey(key, _globally_unique=True))
        return val
    return wrapper


class FlagClearDecorator(ModelingDecoratorWrapper):

    def __init__(self, value, flag: NSFlags):
        super().__init__(value)
        self.flag = flag

    def handle(self, cls, ns, k, state):
        super().handle(cls, ns, k, state)
        state.flags &= ~self.flag


def no_instantiate():
    def wrapper(val):
        return FlagClearDecorator(val, NSFlags.instantiate_on_access)
    return wrapper


def no_close():
    def wrapper(val):
        return FlagClearDecorator(val, NSFlags.close)
    return wrapper


def allow_multiple():
    raise NotImplementedError("Need to write FlagSetDecorator")


class MachineMixin(ModelingDecoratorWrapper):

    subclass = InjectableModelType
    name = "machine_mixin"

    def __init__(self, value, name):
        super().__init__(value)
        self.name = name

    def handle(self, cls, ns, k, state):
        super().handle(cls, ns, k. state)
        state.flags &= ~(NSFlags.inject_by_name | NsFlags.inject_by_class | NSFlags.instantiate_on_access)
        ns.to_inject[InjectionKey(MachineMixin,
                                  name=self.name)] = (
                                      dependency_quote(state.value), state.injection_options)


def machine_mixin(name=None):
    '''Mark a class (subclass of :class:`Machine` typically) as something that should be mixed in to any machine declared lower in the injector hierarchy withing modeling classes.  To accomplish the same thing outside of modeling classes::

        injector.add_provider(InjectionKey(MachineMixin, name = "some_name"), dependency_quote(mixin_class))

    The call to :func:~carthage.dependency_injection.dependency_quote`
    is required to prevent the injector from trying to build a machine
    when the Mixin is looked up.

'''
    def wrapper(val):
        return MachineMixin(val, name or val.__name__)
    return wrapper


class PropagateUpDecorator(ModelingDecoratorWrapper):

    name = "propagate_up"
    subclass = ModelingContainer

    def handle(self, cls, ns, k, state):
        super().handle(cls, ns, k, state)
        state.flags |= NSFlags.propagate_key


def propagate_up():
    '''Indicate that an assignment should have its keys propagated up in
    a container::

        class foo(ModelingContainer):
            @propagate_up()
            @provides(InjectionKey(Baz, target = 42))
            class ourbaz(Baz): ...

    When *foo* is included ain a container, then the *Baz* injection
    key will be propagated up to dependencies provided by that
    container.  Since the key was not marked globally unique,
    constraints from *foo.our_key()* will be added to it as it is
    propagated.

    '''
    def wrap(val):
        return PropagateUpDecorator(val)
    return wrap


def transclude_overrides(
        key: InjectionKey = None):
    '''
    Decorator indicating that the decorated item should be overridden by one from the injector against which the containing layout is eventually instantiated.  When a :class:`InjectableModel` is eventually instantiated, before an overridable key is added to the local injector, it is searched for in the instantiating injector.  If it is found, the  key is not registered.  This has the effect of using the dependency provider in the instantiating injector rather than the one included in the layout.

    :param key: If supplied, is the key expected to be registered with the transcluding injector to override  this object.  If not supplied, the object must have a globally unique key.



    Example usage::

        class layout(CarthageLayout):

            @transclude_overrides(key=InjectionKey("network"))
            class network(NetworkModel):
                # ...

    If ``layout`` is instantiated in a injector that provides a dependency for ``InjectionKey('network')``, then that object will be used rather than the ``network`` class within the layout.  Since the ``network`` property is an :func:`injector_acess`, ``layout_instance.network`` will refer to the object in the instantiating injector rather than an instance of ``layout.network``.


    '''
    def wrap(val):
        nonlocal key
        if key is None:
            for k in val.__provides_dependencies_for__:
                if k.globally_unique:
                    key = k
                    break
            if key is None:
                raise ValueError('A globally unique key is required')
        val.__transclusion_key__ = key
        return val
    return wrap



def model_mixin_for(**constraints):
    '''
    Indicate that a given model supplements a model declared automatically.    The simplest usage looks like::

        @model_mixin_for(host = "foo.com")
        class foomixin(MachineModel):
            # stuff added to foo.com

        class foo(MachineModel):
            name = "foo.com"
            # Also inherits from foomixin

    In the above example it would be simpler to list *foomixin* as a base for *foo*.  However when a loop instantiates a number of models with a dynamic name, *model_mixin_for* provides value::

        for c in ('foo', 'bar', 'baz'):
            @dynamic_name(c)
            class model(MachineModel):
                name = c+".com"

    In the above usage, the loop would need to get more complicated to only add *foomixin* to the dynamically generated class for the ``foo.com`` model.  In this case *model_mixin_for* provides value.

    IN a more complex usage, *model_mixin_for* can be used in a layout that will transclude a model using :func:`carthage.modeling.base.model_bases`.  This is similar to :func:`transclude_overrides` except that rather than entirely replacing the model, *model_mixin_for* simply adds a base.

    '''
    from .base import MachineModelMixin

    def wrap(val):
        if isinstance(val, ModelingDecoratorWrapper):
            # Assume it is already a model_mix in_for which we're stacking
            return provides(InjectionKey(MachineModelMixin, **constraints))(val)
        return provides(InjectionKey(
            MachineModelMixin, **constraints))(dependency_quote(val))
    return wrap


class wrap_base_customization:

    def __init__(self, val, name):
        self.name = name
        functools.wraps(val)(self)
        self.value = val

    def __get__(self, instance, owner):
        if instance is None:
            return self.value
        machine = instance.injector.get_instance(InjectionKey(carthage.machine.Machine, _ready=False))
        return instance.injector(self.value, apply_to=machine, stamp=self.name)


__all__ = ["ModelingDecoratorWrapper", "provides", 'dynamic_name',
           'injector_access', 'no_instantiate',
           'allow_multiple', 'no_close',
           'globally_unique_key',
           'MachineMixin', 'machine_mixin',
           'propagate_up',
           'transclude_overrides',
           'model_mixin_for',
           ]



# Copyright (C) 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import pytest
from carthage.event import EventListener
from carthage.pytest import *
from carthage.dependency_injection import *
from carthage import base_injector


@async_test
async def test_event_register(loop):
    listener = EventListener()
    callback_called = 0

    def callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = 1
    listener.add_event_listener("foo", "event_1", callback)
    listener.loop = loop
    await listener.emit_event("foo", "event_1", listener)
    assert callback_called == 1


@async_test
async def test_event_adl_keys(loop):
    listener = EventListener()
    callback_called = 0

    def callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = 1
    listener.add_event_listener("foo", "event_1", callback)
    listener.loop = loop
    await listener.emit_event("bar", "event_1", listener,
                              adl_keys={'foo'})
    assert callback_called == 1


@async_test
async def test_event_scoping(loop):
    injector = base_injector(Injector)
    injector2 = injector(Injector)
    ainjector = injector2(AsyncInjector)
    key = InjectionKey("baz")
    callback_called = 0

    def callback(**kwargs):
        nonlocal callback_called
        callback_called = 1
    injector.add_event_listener(key, "foo", callback)
    await injector2.emit_event(key, "foo", injector2)
    assert callback_called == 1
    callback_called = 0
    injector2.add_event_listener(key, "bar", callback)
    await injector2.emit_event(key, "foo", injector2)
    assert callback_called == 1


def test_multiple_scope_breaks(loop):
    def callback(*args): pass
    injector = base_injector(Injector).claim("injector")
    injector2 = injector(Injector).claim("injector2")
    injector3 = injector2(Injector).claim("injector3")
    key = InjectionKey("event")
    injector3.add_event_listener(key, "foo", callback)
    injector2.add_event_listener(key, "foo", callback)



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.


def test_true():
    assert True
